from typing import Union

import numpy as np
import pandas as pd
import plotly.graph_objects as go


def plot_ohlc(ohlc_data: dict[str, pd.DataFrame],
			  ohlc_key: Union[int, str] = 0) -> Union[go.Figure, None]:
	"""
	Creates a plotly plot of all ohlc data to be displayed by Dash server. Set ohlc_key to change displayed
	dataframe.

	Args:
		ohlc_data (dict{key: DataFrame}): OHLC Data
		ohlc_key (int/ str): Dictionary key

	Returns:
		Plotly plot
	"""
	# Check
	empty_ohlc = [k for k, v in ohlc_data.items() if v.empty]
	ohlc_data = {k: v for k, v in ohlc_data.items() if not v.empty}
	if not ohlc_data:
		raise ValueError("All ohlc objects are empty!")
	if empty_ohlc:
		print(f"Warning: The following ohlc objects are empty and have been removed: {empty_ohlc}")

	# Setup
	try:
		ohlc_key = [*ohlc_data][ohlc_key] if isinstance(ohlc_key, int) else ohlc_key
	except (IndexError, TypeError):
		ohlc_key = [*ohlc_data][0]

	ohlc_data = ohlc_data[ohlc_key]

	# Plotting ohlc-Diagram with Plotly
	try:
		fig = go.Figure(data=go.Candlestick(x=ohlc_data.index,  # Alternative: go.ohlc()
											open=ohlc_data['open'],
											high=ohlc_data['high'],
											low=ohlc_data['low'],
											close=ohlc_data['close']))

		fig.update_layout(title='OHLC-Chart', yaxis_title='Price per MWh')
		return fig
	except:
		return None


def plot_orderbook(orderbooks: dict[str, dict[str, pd.DataFrame]],
				   orderbook_key: Union[int, str] = 0,
				   timestamp: Union[int, str] = -1) -> Union[go.Figure, None]:
	"""
		Creates a plotly plot of a single order book to be displayed by Dash server. Use order book_key to specify an
		order book and timestamp to specify the timeframe to display.

	Args:
		orderbooks (dict{key: DataFrame}): Order books
		orderbook_key (int): Dictionary key
		timestamp (int): Order book Key

	Returns:
		Plotly plot
	"""
	# Check
	empty_ob = [k for k, v in orderbooks.items() if isinstance(v, pd.DataFrame) or not v]
	orderbooks = {k: v for k, v in orderbooks.items() if not isinstance(v, pd.DataFrame) and v}
	if not orderbooks:
		raise ValueError("All order books are empty!")
	if empty_ob:
		print(f"Warning: The following order books are empty and have been removed: {empty_ob}")

	# Setup
	line_shape = 'hv'
	mode = 'lines+markers'
	try:
		orderbook_key = [*orderbooks][orderbook_key] if isinstance(orderbook_key, int) else orderbook_key
	except (IndexError, TypeError):
		orderbook_key = [*orderbooks][0]

	try:
		timestamp = [*orderbooks[orderbook_key]][timestamp] if isinstance(timestamp, int) else timestamp
	except (IndexError, TypeError):
		timestamp = [*orderbooks][0][-1]

	orders = {"bid": {"name": "Bids", "sort_order": [True, False]}, "ask": {"name": "Asks", "sort_order": [True, True]}}

	# Plotting With Plotly
	df_plot = orderbooks[orderbook_key][timestamp].sort_values(by=['price'], ascending=True)
	fig = go.Figure()

	try:
		for i in ["bid", "ask"]:
			prices = df_plot.price[df_plot.type == i].to_list()
			if len(prices) == 0:
				continue
			quantities = df_plot.quantity[df_plot.type == i].to_list()

			# Add Quantities
			if i == "bid":
				quantities_added = np.cumsum(quantities[::-1])[::-1]
			else:
				quantities_added = np.cumsum(quantities)

			temp_dict = {}
			for x, y in enumerate(prices):
				temp_dict[x] = [y, quantities_added[x]]

			df_temp_plot = pd.DataFrame(temp_dict).T.sort_values(by=[0, 1], ascending=orders[i]["sort_order"])
			df_temp_plot.columns = ["price", "quantity"]

			fig.add_trace(
				go.Scatter(x=df_temp_plot.price, y=df_temp_plot.quantity, name=orders[i]["name"], fill='tozeroy',
						   line_shape=line_shape, mode=mode))  # fill down to xaxis

		fig.update_layout(title=f'Orderbook {orderbook_key} - {timestamp.split(" ")[1]}',
						  xaxis_title="Price per MWh", yaxis_title='Quantity')

	except:
		fig = go.Figure()
		fig.add_trace(go.Scatter())
		fig.update_layout(title=f'Orderbook {orderbook_key}', xaxis_title="Price per MWh", yaxis_title='Quantity')

	return fig


def ohlc_table(ohlc_data: dict[str, pd.DataFrame], ohlc_key: int = 0):
	"""
	Creates a custom DataFrame to be displayed by Dash server.

	Args:
		ohlc_data (dict{key: DataFrame}): OHLC Data
		ohlc_key (int): Dictionary key

	Returns:
		DataFrame
	"""
	# Check
	empty_ohlc = [k for k, v in ohlc_data.items() if v.empty]
	ohlc_data = {k: v for k, v in ohlc_data.items() if not v.empty}
	if not ohlc_data:
		raise ValueError("All ohlc objects are empty!")
	if empty_ohlc:
		print(f"Warning: The following ohlc objects are empty and have been removed: {empty_ohlc}")

	# Setup
	try:
		ohlc_key = [*ohlc_data][ohlc_key]
	except (IndexError, TypeError):
		ohlc_key = [*ohlc_data][0]

	ohlc_contract_data = ohlc_data[ohlc_key].reset_index()
	ohlc_contract_data = ohlc_contract_data.rename(columns={"index": "Timestamp"})

	return ohlc_contract_data


def plot_volume_history(trade_data: dict[str, pd.DataFrame],
						trade_key: Union[int, str] = 0) -> Union[go.Figure, None]:
	"""
	Creates a plotly plot of the trade volume for a single contract to be displayed by Dash server.

	Args:
		trade_data (dict{key: DataFrame}):  Trade Data
		trade_key (int/str): Dictionary key

	Returns:
		Plotly plot
	"""
	# Check
	empty_trades = [k for k, v in trade_data.items() if v.empty]
	trade_data = {k: v for k, v in trade_data.items() if not v.empty}
	if not trade_data:
		raise ValueError("There are no trades!")
	if empty_trades:
		print(f"Warning: The following trade collections are empty and have been removed: {empty_trades}")

	line_shape = 'hv'
	mode = 'lines+markers'
	try:
		trade_key = [*trade_data][trade_key] if isinstance(trade_key, int) else trade_key
	except (IndexError, TypeError):
		trade_key = [*trade_data][0]

	df_trade = trade_data[trade_key].sort_values(by=["exec_time"], ascending=[True])
	quantities = np.cumsum(df_trade["quantity"].tolist())

	fig = go.Figure()
	fig.add_trace(
		go.Scatter(x=df_trade.exec_time, y=quantities, name="trades", fill='tozeroy', line_shape=line_shape, mode=mode,
				   line_color='rgb(34,139,34)'))  # fill down to xaxis
	fig.update_layout(title="Trade Volume", xaxis_title="Time", yaxis_title='Quantity')

	return fig



import asyncio
import gzip
import json
import pickle
from datetime import datetime, timedelta, timezone
from typing import Union

import numpy as np
import pandas as pd
from powerbot_client import ApiClient, ContractApi

from powerbot_backtesting.data_acquisition import get_public_trades_by_days
from powerbot_backtesting.exceptions import NotInCacheError
from powerbot_backtesting.models import HistoryApiClient
from powerbot_backtesting.utils import _find_cache, _cache_data, _orderbook_data_transformation, _delta_filter
from powerbot_backtesting.utils.constants import *


def get_orders(contract_hist_data: dict[str, pd.DataFrame]) -> dict[str, list[dict]]:
	"""
	Extracts all order data from contract history as is, without performing any quality control.

	Args:
		contract_hist_data (dict): Dictionary of Dataframes containing Contract History Data

	Returns:
		dict{key: [orders]}: Dictionary of Lists of Orders
	"""
	order_list = {}

	for key, value in contract_hist_data.items():
		value.replace(np.nan, 0, inplace=True)
		bids_asks = []
		if "orders" in value:
			orders_all = value["orders"].to_list()
			for nr, orders in enumerate(orders_all):
				for k, v in orders.items():
					if v and k in ["bid", "ask"]:
						for x in v:
							x["type"] = k
							x["best_bid"] = round(value.loc[nr, "best_bid_price"], 2)
							x["best_bid_qty"] = round(value.loc[nr, "best_bid_quantity"], 2)
							x["best_ask"] = round(value.loc[nr, "best_ask_price"], 2)
							x["best_ask_qty"] = round(value.loc[nr, "best_ask_quantity"], 2)
							try:
								x["vwap"] = round(value.loc[nr, "vwap"], 2)
							except:
								pass
							x["as_of"] = value.loc[nr]["as_of"].tz_convert(timezone.utc) if value.loc[nr]["as_of"].tzinfo else \
							value.loc[nr]["as_of"].tz_localize(timezone.utc)
							bids_asks.append(x)
		else:
			value = [v.to_dict() for r, v in value.iterrows()]
			for nr, coll in enumerate(value):
				for k, v in coll.items():
					if v and k in ["bids", "asks"]:
						for x in v:
							x["type"] = k[:-1]
							x["best_bid"] = round(value[nr].get("best_bid", 0), 2)
							x["best_bid_qty"] = round(value[nr].get("best_bid_qty", 0), 2)
							x["best_ask"] = round(value[nr].get("best_ask", 0), 2)
							x["best_ask_qty"] = round(value[nr].get("best_ask_qty", 0), 2)
							x["vwap"] = round(value[nr].get("vwap", 0), 2)
							x["as_of"] = value[nr]["as_of"].tz_convert(timezone.utc) if value[nr]["as_of"].tzinfo else \
								value[nr]["as_of"].tz_localize(timezone.utc)
							bids_asks.append(x)
							bids_asks.append(x)

		order_list[key] = bids_asks

	return order_list


def get_ohlc_data(trade_data: dict[str, pd.DataFrame],
                  timesteps: int,
                  time_unit: str,
                  delivery_area: str = None,
                  api_client: Union[ApiClient, HistoryApiClient] = None,
                  use_cached_data: bool = True,
                  caching: bool = True,
                  gzip_files: bool = True,
                  one_file: bool = False) -> dict[str, pd.DataFrame]:
	"""
	Converts trade data into Open-High-Low-Close format in the specified timesteps.

	Args:
		trade_data (dict{key: DataFrame}): Dictionary of Dataframes containing Contract Trade Data
		timesteps (int): Timesteps to group Trades by
		time_unit (str): Time units for timesteps (either hours, minutes or seconds)
		delivery_area (str): Area Code for Delivery Area (not needed when loading from historic cache)
		api_client: PowerBot ApiClient
		use_cached_data (bool): If True, function tries to load data from cache wherever possible
		caching (bool): True if data should be cached
		gzip_files (bool): True if cached files should be gzipped
		one_file (bool): True if data should be cached in a single JSON file

	Returns:
		dict{key: DataFrame}: Dictionary of DataFrames
	"""
	# Setup Parameters
	all_ohlc_data = {}
	host = api_client.configuration.host if isinstance(api_client, ApiClient) else None
	environment = host.split("/")[2].split(".")[0] if host else "prod"
	exchange = host.split("/")[4] if host else api_client.exchange if isinstance(api_client, HistoryApiClient) else \
	list(trade_data.values())[0].exchange[0]
	file_ending = ".gz" if gzip_files else ""
	delivery_area = api_client.delivery_area if isinstance(api_client, HistoryApiClient) else delivery_area

	# Check if __cache__ already exists
	cache_path = _find_cache()

	for key, value in trade_data.items():
		# Check If Data Already Cached
		delivery_date = datetime.strptime(key.split(" ")[0], DATE_YMD)
		year_month = delivery_date.strftime(DATE_YM)
		day_month = delivery_date.strftime(DATE_MD)
		file_name = key.replace(f"{str(key).split(' ')[0]}", "").replace(":", "-")
		data_ohlc = None

		if use_cached_data:
			for i in [".json.gz", ".json"]:
				if cache_path.joinpath(f'{environment}\\{exchange}_{delivery_area}\\{year_month}\\{day_month}'
				                       f'\\processed\\{file_name}_ohlc_{timesteps}{time_unit[0]}{i}').exists():
					data_ohlc = pd.read_json(cache_path.joinpath(f'{environment}\\{exchange}_{delivery_area}\\'
					                                             f'{year_month}\\{day_month}\\processed\\'
					                                             f'{file_name}_ohlc_{timesteps}{time_unit[0]}{i}'),
					                         dtype=False)

		if isinstance(data_ohlc, pd.DataFrame):
			data_ohlc.rename(columns={0: 'exec_time'}, inplace=True)
			all_ohlc_data[key] = data_ohlc

		else:
			data_ohlc = value.set_index('exec_time')
			data_ohlc = data_ohlc['price'].resample(f'{timesteps}{time_unit[0]}').ohlc() if time_unit != "minutes" \
				else data_ohlc['price'].resample(f'{timesteps}{time_unit[:3]}').ohlc()
			data_ohlc = data_ohlc.dropna(how='all')

			# Append to complete OHLC collection
			all_ohlc_data[key] = data_ohlc

	# Cache Data as JSON
	if caching:
		_cache_data("ohlc", all_ohlc_data, delivery_area, exchange=exchange, api_client=api_client,
		            gzip_files=gzip_files, timesteps=timesteps, time_unit=time_unit[0], as_csv=False)

	# Cache Data as JSON
	if caching:
		_cache_data("ohlc", all_ohlc_data, delivery_area, exchange=exchange, api_client=api_client,
		            gzip_files=gzip_files, timesteps=timesteps, time_unit=time_unit[0], as_csv=False)

	# Saving Complete OHLC Data as JSON
	if one_file and trade_data:
		# Take Day of Delivery Start
		# Parameters
		contract_times = sorted([i for i in [*trade_data]])
		first_contract = datetime.strptime(
			contract_times[0].replace(f"{str(contract_times[0]).split(' - ')[1]}", "").replace(" - ", ":00"),
			DATE_YMD_TIME_HMS)
		last_contract = datetime.strptime(
			contract_times[-1].replace(f"{str(contract_times[-1]).split(' ')[1]}", "").replace(" - ", "") + ":00",
			DATE_YMD_TIME_HMS)

		year_month = first_contract.strftime(DATE_YM)
		day_month = first_contract.strftime(DATE_MD)

		if len(contract_times) == 1:
			first_contract = first_contract.strftime(DATE_YMD_TIME_HMS_ALT)
			last_contract = None
		else:
			if first_contract.strftime(DATE_D) != last_contract.strftime(DATE_D):
				first_contract = first_contract.strftime(DATE_YMD_TIME_HM_ALT)
				last_contract = last_contract.strftime(DATE_YMD_TIME_HM_ALT)
			else:
				first_contract = first_contract.strftime(TIME_HM_ALT)
				last_contract = last_contract.strftime(TIME_HM_ALT)

		filename = f'all_ohlc_{first_contract} - {last_contract}_{timesteps}{time_unit[0]}.json.gz' \
			if last_contract else f'all_ohlc_{first_contract}_{timesteps}{time_unit[0]}.json{file_ending}'

		if not cache_path.joinpath(
				f"{environment}\\{exchange}_{delivery_area}\\{year_month}\\{day_month}\\processed\\{filename}").exists():
			if "gz" in file_ending:
				with gzip.open(cache_path.joinpath(
						f"{environment}\\{exchange}_{delivery_area}\\{year_month}\\{day_month}\\processed\\{filename}"),
						'wt', encoding="ascii") as f:
					json.dump(
						{key: {k: v.to_json() for (k, v) in value.items()} for (key, value) in all_ohlc_data.items()},
						f, default=str)
			else:
				with open(cache_path.joinpath(
						f"{environment}\\{exchange}_{delivery_area}\\{year_month}\\"
						f"{day_month}\\processed\\{filename}"), 'wt',
						encoding="ascii") as f:
					json.dump(
						{key: {k: v.to_json() for (k, v) in value.items()} for (key, value) in all_ohlc_data.items()},
						f, default=str)

	return all_ohlc_data


def get_orderbooks(contract_hist_data: dict[str, pd.DataFrame],
                   delivery_area: str = None,
                   timesteps: int = 15,
                   time_unit: str = "minutes",
                   timestamp: list[datetime] = None,
                   from_timestamp: bool = False,
                   api_client: Union[ApiClient, HistoryApiClient] = None,
                   use_cached_data: bool = True,
                   caching: bool = True,
                   as_json: bool = False) -> dict[str, pd.DataFrame]:
	"""
	Converts contract history data into order books in the specified timesteps. If no API client is passed, the function
	will automatically assume that the data is production data.

	Please be aware that optimally only data from one exchange at a time should be used (e.g. only EPEX).

	To generate specific order books for a position closing algorithm, the timestamp and from_timestamp parameters can
	be used.

	Args:
		contract_hist_data (dict{key: DataFrame}): Dictionary of Dataframes containing Contract History Data
		delivery_area (str): Area Code for Delivery Area (not needed when loading from historic cache)
		timesteps (int): Timesteps to group order books by
		time_unit (str): Time units for timesteps (either hours, minutes or seconds)
		timestamp (list[datetime]): List of timestamps to generate order books at/ from
		from_timestamp (bool): True if timestamp serves as starting point for order book generation
		api_client: PowerBot ApiClient
		use_cached_data (bool): If True, function tries to load data from cache wherever possible
		caching (bool): True if single order books should be cached as JSON
		as_json (bool): True if complete order book should be cached as JSON

	Returns:
		dict{key: DataFrame}: Dictionary of DataFrames
	"""
	# Initial check
	if not contract_hist_data:
		raise ValueError("Warning: Provided order data is empty")

	# Setup
	all_order_books = {}
	host = api_client.configuration.host if isinstance(api_client, ApiClient) else None
	environment = host.split("/")[2].split(".")[0] if host else "prod"
	exchange = host.split("/")[4] if host else api_client.exchange if isinstance(api_client, HistoryApiClient) else \
	list(contract_hist_data.values())[0].exchange[0]
	delivery_area = api_client.delivery_area if isinstance(api_client, HistoryApiClient) else delivery_area

	# Parameters
	contract_times = sorted([i for i in [*contract_hist_data]])
	first_contract = datetime.strptime(
		contract_times[0].replace(f"{str(contract_times[0]).split(' - ')[1]}", "").replace(" - ", ":00"),
		DATE_YMD_TIME_HMS)
	last_contract = datetime.strptime(
		contract_times[-1].replace(f"{str(contract_times[-1]).split(' ')[1]}", "").replace(" - ", "") + ":00",
		DATE_YMD_TIME_HMS)

	year_month = first_contract.strftime(DATE_YM)
	day_month = first_contract.strftime(DATE_MD)

	if len(contract_times) == 1:
		first_contract = first_contract.strftime(DATE_YMD_TIME_HM_ALT)
		last_contract = None
	else:
		if first_contract.strftime(DATE_D) != last_contract.strftime(DATE_D):
			first_contract = first_contract.strftime(DATE_YMD_TIME_HM)
			last_contract = last_contract.strftime(DATE_MD_TIME_HM)
		else:
			first_contract = first_contract.strftime(TIME_HM_ALT)
			last_contract = last_contract.strftime(TIME_HM_ALT)

	# Check if __cache__ already exists
	cache_path = _find_cache()
	new_dir = cache_path.joinpath(f"{environment}\\{exchange}_{delivery_area}\\{year_month}\\{day_month}\\processed")

	# Processing Concurrently With Asyncio
	async def main():
		function_list = []
		for nr, (key, value) in enumerate(contract_hist_data.items()):
			function_list.append(
				__process_orderbook(key, value, str(new_dir), timesteps, time_unit,
				                    timestamp[nr] if timestamp else None, from_timestamp,
				                    all_order_books, use_cached_data))

		await asyncio.gather(*function_list)

	asyncio.run(main())

	# Cache Data as pickle
	if caching:
		_cache_data("orderbook", all_order_books, delivery_area, exchange=exchange, api_client=api_client,
		            gzip_files=False, timesteps=timesteps, time_unit=time_unit[0], as_json=False, as_pickle=True)

	# Saving Complete Orderbook as JSON
	if as_json:
		new_dir.mkdir(parents=True, exist_ok=True)

		filename = f'orderbook_{first_contract} - {last_contract}_{timesteps}{time_unit[0]}.json.gz' \
			if last_contract else f'orderbook_{first_contract}_{timesteps}{time_unit[0]}.json.gz'
		with gzip.open(new_dir.joinpath(filename), 'wt', encoding="ascii") as f:
			json.dump({key: {k: v.to_json() for (k, v) in value.items()} for (key, value) in all_order_books.items()},
			          f, default=str)

	return all_order_books


async def __process_orderbook(key: str,
                              value: pd.DataFrame,
                              directory: str,
                              timesteps: int,
                              time_unit: str,
                              timestamp: Union[datetime, None],
                              from_timestamp: bool,
                              orderbook_dict: dict[str, pd.DataFrame],
                              use_cached_data: bool):
	"""
	Asyncio function to process single order book. Return value is appended to collection of order books.

	Returns:
		pd.Dataframe: single order book
	"""
	# Setup Parameters
	units = {"hours": 0, "minutes": 0, "seconds": 0, time_unit: timesteps}
	delivery_start = datetime.strptime(key.replace(f"{str(key).split(' - ')[1]}", "").replace(" - ", ":00"),
	                                   DATE_YMD_TIME_HMS).replace(tzinfo=timezone.utc)
	timestamp = timestamp.replace(tzinfo=timezone.utc) if timestamp else None
	file_name = key.replace(f"{str(key).split(' ')[0]}", "").replace(":", "-")
	directory = directory.split('\\')
	directory = '\\'.join(directory)

	try:
		if not use_cached_data:
			raise NotInCacheError("Not loading from cache")
		# Check If Data Already Cached
		order_book_clean = pickle.load(open(f"{directory}\\{file_name}_orderbook_{timesteps}{time_unit[0]}.p", "rb"))
		orderbook_dict[key] = order_book_clean

	except (NotInCacheError, FileNotFoundError):
		# Filter out emtpy revisions
		if "orders" in value:
			order_filter = value.orders.map(lambda x: x if x["bid"] or x["ask"] else None)
			value = value.loc[~value.index.isin(order_filter[order_filter.isna()].index)]
		else:
			value = value.loc[~(value.bids.isna()) & ~(value.asks.isna())]

		# Setting Either Starting Point or Specific Timestamp
		start_time = value.as_of.min().replace(minute=0, second=0, microsecond=0, tzinfo=timezone.utc)
		time = timestamp if timestamp and from_timestamp \
			else timestamp + timedelta(**{time_unit: timesteps}) if timestamp and not from_timestamp \
			else start_time + timedelta(**{time_unit: timesteps})
		time = start_time + timedelta(**{time_unit: timesteps}) if start_time > time else time

		order_book = {}

		# Transform orders
		df_bid_asks = _orderbook_data_transformation(value)

		if not df_bid_asks.empty:
			# Create a filter to shorten dataframe
			orders_del = set()

			# Main Loop
			while time <= delivery_start:
				# Shorten Bids_Asks
				if len(orders_del) != 0:
					df_bid_asks = df_bid_asks.loc[~df_bid_asks.order_id.isin(orders_del)]

				# Create New Temporary Dataframe
				if timestamp and not from_timestamp:
					df_temp = df_bid_asks.loc[df_bid_asks.as_of <= f'{timestamp}']

				else:
					df_temp = df_bid_asks.loc[(df_bid_asks.as_of >= f'{start_time}') & (df_bid_asks.as_of <= f'{time}')]

				# Delete all orders before the last delta == False
				df_temp = _delta_filter(df_temp)

				if not df_temp.empty:
					# Extract Order IDs for Quantity = 0 & Update Set Of Order IDs
					contract_ids = df_temp.contract_id.unique().tolist()
					orders_del.update(df_temp.loc[df_temp.quantity == 0].order_id.tolist())

					# Check For Uniformity of Contract ID
					if len(contract_ids) == 1:
						# QC For Temporary Dataframe
						# Add Filtered Df To Orderbook
						if timestamp and not from_timestamp:
							order_book[f"{timestamp}"] = df_temp.loc[~df_temp.order_id.isin(orders_del)]
						else:
							order_book[f"{time}"] = df_temp.loc[~df_temp.order_id.isin(orders_del)]

					else:
						# If There Are Multiple Contracts In The Same Orderbook -> Create 2 Separate Orderbooks
						dataframes = []
						df_check_1 = df_temp.loc[df_temp.contract_id == contract_ids[0]]
						dataframes.append(df_check_1)
						df_check_2 = df_temp.loc[df_temp.contract_id == contract_ids[1]]
						df_check_2 = df_check_2.loc[df_check_2.as_of > f'{time - timedelta(**units)}']
						dataframes.append(df_check_2)

						# Quality Control For Temporary Dataframe
						temp_dataframe_list = []

						for nr, val in enumerate(dataframes):
							# Quality Control For Temporary Dataframe
							df_check = val.loc[~val.order_id.isin(orders_del)]

							if not df_check.empty:
								temp_dataframe_list.append(df_check)  # Add Filtered Df To List

						if len(temp_dataframe_list) == 1:
							order_book[f"{timestamp if timestamp and not from_timestamp else time}"] = \
								temp_dataframe_list[0]
						else:
							for nr, val in enumerate(temp_dataframe_list):
								if not nr:
									order_book[
										f"{(timestamp if timestamp and not from_timestamp else time) - timedelta(seconds=1)}"] = val
								else:
									order_book[f"{timestamp if timestamp and not from_timestamp else time}"] = val

				# Progress In Time Or Break Loop If Timestamp Exists
				if timestamp and not from_timestamp:
					break

				# Adjust Start Time To New Contract if necessary
				if time == (delivery_start - timedelta(hours=1)):
					start_time = time

				# Time Progression
				time += timedelta(**units)

			# General Quality Control
			# Delete All Order ID Duplicates & Empty Timesteps
			order_book_clean = {
				key: value.sort_values(by=['as_of'], ascending=False).drop_duplicates(subset="order_id", keep="first",
				                                                                      inplace=False) for (key, value) in
				order_book.items()}
			order_book_clean = {key: value for (key, value) in order_book_clean.items() if not value.empty}

		else:
			order_book_clean = df_bid_asks

	orderbook_dict[key] = order_book_clean


def calc_trade_vwap(api_client: ApiClient,
                    contract_time: str,
                    delivery_area: str,
                    trade_data: dict[str, pd.DataFrame] = None,
                    time_from: datetime = None,
                    previous_days: int = 10,
                    contract_id: str = None,
                    index: str = "ID3") -> pd.DataFrame:
	"""
	Function gets trades for a certain contract for X previous days in the same delivery period and calculates their
	VWAP for ID3 or ID1 or all. Generates a new list of trades for these contracts.
	Can take either a time period or a specific contract ID to load data for.

	If previous days is 0, only the trades for the original time period/ contract will be loaded.

	Can also be called directly from get_public_trades with parameter 'add_vwap' to add VWAP to loaded trades.

	Args:
		api_client: PowerBot ApiClient
		contract_time (str): hourly, half-hourly or quarter-hourly
		delivery_area (str): Area Code for Delivery Area
		trade_data (dict[str, pd.DataFrame]: Dictionary of Dataframes containing Contract Trade Data
		time_from (str/ datetime): yyyy-mm-dd hh:mm:ss
		previous_days (int): Amount of previous days to load data
		contract_id (str): ID of specific Contract
		index (str): all, ID3, ID1

	Returns:
		DataFrame: Trade Data with added calculated fields
	"""
	if not trade_data and contract_time == "all":
		raise ValueError("Contract time can only be one of hourly, half-hourly or quarter-hourly if no trade data is given")

	# Setup
	indices = {"all": 1980, "ID3": 180, "ID1": 60}
	contract_api = ContractApi(api_client)

	# Create Empty Dataframe
	all_trade_data = pd.DataFrame(
		{'trade_id': [], 'buy_delivery_area': [], 'sell_delivery_area': [], 'api_timestamp': [],
		 'exec_time': [], 'contract_id': [], 'price': [], 'quantity': [], 'self_trade': [], 'time_diff': []})

	# Get Delivery Start If Contract ID was passed
	if contract_id:
		if not isinstance(contract_id, str):
			raise TypeError("contract_id has to be a string")
		time_from = str(contract_api.find_contracts(contract_id=[contract_id])[0].delivery_start).replace("+00:00", "")

	# Get Trade Data
	trade_data = trade_data if trade_data else get_public_trades_by_days(api_client=api_client,
	                                                                     time_from=time_from,
	                                                                     previous_days=previous_days,
	                                                                     delivery_area=delivery_area,
	                                                                     contract_time=contract_time)

	# Processing
	for key, value in trade_data.items():
		# Time Difference In Minutes
		time = datetime.strptime(key.replace(f"{str(key).split(' - ')[1]}", "").replace(" - ", ":00"),
		                         DATE_YMD_TIME_HMS)
		time_diff = [round((time.replace(tzinfo=None) - i.replace(tzinfo=None)).total_seconds() / 60, 2) for i in
		             value.exec_time]
		value["time_diff"] = time_diff
		all_trade_data = all_trade_data.append(value)

	all_trade_data = all_trade_data.loc[all_trade_data["time_diff"] <= indices[index]].sort_values(by=['time_diff'],
	                                                                                               ascending=False)
	total_quantity = all_trade_data.quantity.sum()
	all_quantities = all_trade_data.quantity.tolist()
	all_prices = all_trade_data.price.tolist()

	target_volume = []
	cumulated_quantities = []
	calculated_vwaps = []
	cum_weighted_price = 0

	for nr, item in enumerate(all_quantities):
		cum_sum = round(sum(all_quantities[:nr + 1]), 2)
		cum_weighted_price += all_prices[nr] * all_quantities[nr]
		calculated_vwaps.append(round(cum_weighted_price / cum_sum, 2))
		cumulated_quantities.append(cum_sum)
		target_volume.append(round(cum_sum / total_quantity, 4))

	all_trade_data["cumulated_quantity"] = cumulated_quantities
	all_trade_data["target_volume"] = target_volume
	all_trade_data["vwap"] = calculated_vwaps

	return all_trade_data.reset_index(drop=True)


def calc_single_vwap(objects: dict[str, pd.DataFrame],
                     desired_depth: float,
                     min_depth: float = None) -> dict[str, float]:
	"""
	This method can be used to calculate the weighted average price for a dictionary of dataframes (e.g. orders, trades)
	at a desired depth. The output is a singular value for each dataframe.
	This function does not load any data, therefore the already existing data object has to be passed as an argument.

	Args:
		objects (dict[str, DataFrame or list[str, dict]): A dictionary of dataframes, each of which needs to have a
		'quantity' and a 'price' field. Alternatively, a list of dictionaries can be passed as well.
		desired_depth (float): The depth (in MW) specifying how many of the objects should be taken into consideration.
		min_depth (float): The required minimum depth (in percent of the desired depth). If this requirement is not met,
		return value is 0.

	Returns:
		dict[str, float]: The weighted average price for the desired depth for each key in the dictionary.
	"""
	if min_depth and min_depth > 0.99:
		raise Exception("The minimum depth has to be given as percentage of the desired depth.")

	vwaps = {k: 0 for k in [*objects]}

	for key, obj in objects.items():
		obj = pd.DataFrame(obj) if not isinstance(obj, pd.DataFrame) else obj
		available_depth = 0
		total_value = 0

		for ind, row in obj.iterrows():
			if available_depth + row.quantity < desired_depth:
				available_depth = available_depth + row.quantity
				total_value += row.quantity * row.price
			else:
				total_value += (desired_depth - available_depth) * row.price
				available_depth += desired_depth - available_depth
				available_depth = round(available_depth, 2)
				break

		# If the 'min_depth' parameter is set,
		# then the available depth on the market has to fulfill the minimum requirements.
		if min_depth and available_depth and available_depth > desired_depth * min_depth \
				or not min_depth and available_depth:
			vwaps[key] = round(total_value / available_depth, 2)

	return vwaps



class SQLExporterError(Exception):
    '''
    Called on error during export from database
    '''
    pass

class NotInCacheError(FileNotFoundError):
    '''
    Called when a file has not been found in the local __pb_cache__
    '''
    pass

class ChecksumError(ValueError):
    '''
    Called when the checksum of downloaded files does not match checksum of API
    '''
    pass


from powerbot_backtesting.utils.constants import *
from powerbot_backtesting.utils.helper_functions import _order_matching, _cache_data, _check_contracts, \
	_get_file_cachepath, _find_cache, _orderbook_data_transformation, _historic_data_transformation, _get_private_data, \
	_delta_filter, _historic_contract_transformation



import json
import logging
import re
import urllib.parse
from abc import ABC, abstractmethod
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Union, Optional, TypeVar, Any

import pandas as pd
import plotly.graph_objects as go
from powerbot_client import Signal, Trade, InternalTrade, OwnOrder
from pydantic import BaseModel, Field, validate_arguments, root_validator, validator
from sqlalchemy import create_engine
from sqlalchemy.exc import NoSuchModuleError, InterfaceError, ProgrammingError, OperationalError, TimeoutError
from sqlalchemy.orm import Session

import powerbot_backtesting as pb
from powerbot_backtesting.exceptions import SQLExporterError
from powerbot_backtesting.utils.constants import *

# Implement custom type for validation
pandas_DataFrame = TypeVar('pandas.core.frame.DataFrame')
ApiClient = TypeVar('powerbot_client.api_client.ApiClient')
HistoryApiClient = TypeVar('powerbot_backtesting.models.history_api_models.HistoryApiClient')

class BaseExporter(BaseModel, ABC):
    """
    Base class to all of PowerBot's exporter classes.
    """

    @abstractmethod
    def get_contract_ids(self, **kwargs) -> dict[str, list[str]]:
        """Acquire contracts from source as defined by specific exporter class."""

    @abstractmethod
    def get_public_trades(self, **kwargs) -> dict[str, pd.DataFrame]:
        """Acquire public trades from source as defined by specific exporter class."""

    @abstractmethod
    def get_contract_history(self, **kwargs) -> dict[str, pd.DataFrame]:
        """Acquire contract history from source as defined by specific exporter class."""

    @validate_arguments
    def get_ohlc_data(self,
                      trade_data: dict[str, pandas_DataFrame],
                      timesteps: int,
                      time_unit: str,
                      delivery_area: str = None,
                      use_cached_data: bool = True,
                      caching: bool = True,
                      gzip_files: bool = True,
                      one_file: bool = False) -> dict[str, pd.DataFrame]:
        """
        Converts trade data into Open-High-Low-Close format in the specified timesteps.

        Args:
            trade_data (dict{key: DataFrame}): Dictionary of Dataframes containing Contract Trade Data
            timesteps (int): Timesteps to group Trades by
            time_unit (str): Time units for timesteps (either hours, minutes or seconds)
            delivery_area (str): Area Code for Delivery Area (not needed when loading from historic cache)
            use_cached_data (bool): If True, function tries to load data from cache wherever possible
            caching (bool): True if data should be cached
            gzip_files (bool): True if cached files should be gzipped
            one_file (bool): True if data should be cached in a single JSON file

        Returns:
            dict{key: DataFrame}: Dictionary of DataFrames
        """
        return pb.get_ohlc_data(trade_data=trade_data,
                                timesteps=timesteps,
                                time_unit=time_unit,
                                delivery_area=delivery_area,
                                api_client=self.client if hasattr(self, 'client') else None,
                                use_cached_data=use_cached_data,
                                caching=caching,
                                gzip_files=gzip_files,
                                one_file=one_file)

    @validate_arguments
    def get_orderbooks(self,
                       contract_hist_data: dict[str, pandas_DataFrame],
                       delivery_area: str = None,
                       timesteps: int = 15,
                       time_unit: str = "minutes",
                       timestamp: list[datetime] = None,
                       from_timestamp: bool = False,
                       use_cached_data: bool = True,
                       caching: bool = True,
                       as_json: bool = False) -> dict[str, pd.DataFrame]:
        """
        Converts contract history data into order books in the specified timesteps. If no API client is passed, the function
        will automatically assume that the data is production data.

        Please be aware that optimally only data from one exchange at a time should be used (e.g. only EPEX).

        To generate specific order books for a position closing algorithm, the timestamp and from_timestamp parameters can
        be used.

        Args:
            contract_hist_data (dict{key: DataFrame}): Dictionary of Dataframes containing Contract History Data
            delivery_area (str): Area Code for Delivery Area (not needed when loading from historic cache)
            timesteps (int): Timesteps to group order books by
            time_unit (str): Time units for timesteps (either hours, minutes or seconds)
            timestamp (list[datetime]): List of timestamps to generate order books at/ from
            from_timestamp (bool): True if timestamp serves as starting point for order book generation
            use_cached_data (bool): If True, function tries to load data from cache wherever possible
            caching (bool): True if single order books should be cached as JSON
            as_json (bool): True if complete order book should be cached as JSON

        Returns:
            dict{key: DataFrame}: Dictionary of DataFrames
        """
        return pb.get_orderbooks(contract_hist_data=contract_hist_data,
                                 delivery_area=delivery_area,
                                 timesteps=timesteps,
                                 time_unit=time_unit,
                                 timestamp=timestamp,
                                 from_timestamp=from_timestamp,
                                 api_client=self.client if hasattr(self, 'client') else None,
                                 use_cached_data=use_cached_data,
                                 caching=caching,
                                 as_json=as_json)

    @staticmethod
    @validate_arguments
    def get_orders(contract_hist_data: dict[str, pandas_DataFrame]) -> dict[str, list[dict]]:
        """
        Extracts all order data from contract history as is, without performing any quality control.

        Args:
            contract_hist_data (dict): Dictionary of Dataframes containing Contract History Data

        Returns:
            dict{key: [orders]}: Dictionary of Lists of Orders
        """
        return pb.get_orders(contract_hist_data=contract_hist_data)

    @staticmethod
    @validate_arguments
    def calc_single_vwap(objects: dict[str, pandas_DataFrame],
                         desired_depth: float,
                         min_depth: float = None) -> dict[str, float]:
        """
        This method can be used to calculate the weighted average price for a dictionary of dataframes (e.g. orders, trades)
        at a desired depth. The output is a singular value for each dataframe.
        This function does not load any data, therefore the already existing data object has to be passed as an argument.

        Args:
            objects (dict[str, DataFrame or list[str, dict]): A dictionary of dataframes, each of which needs to have a
            'quantity' and a 'price' field. Alternatively, a list of dictionaries can be passed as well.
            desired_depth (float): The depth (in MW) specifying how many of the objects should be taken into consideration.
            min_depth (float): The required minimum depth (in percent of the desired depth). If this requirement is not met,
            return value is 0.

        Returns:
            dict[str, float]: The weighted average price for the desired depth for each key in the dictionary.
        """
        return pb.calc_single_vwap(objects=objects,
                                   desired_depth=desired_depth,
                                   min_depth=min_depth)

    @staticmethod
    @validate_arguments
    def plot_ohlc(ohlc_data: dict[str, pandas_DataFrame],
                  ohlc_key: Union[int, str] = 0) -> Union[go.Figure, None]:
        """
        Creates a plotly plot of all ohlc data to be displayed by browser or Dash server. Set ohlc_key to change
        displayed dataframe.

        Args:
            ohlc_data (dict{key: DataFrame}): OHLC Data
            ohlc_key (int/ str): Dictionary key

        Returns:
            Plotly plot
        """
        return pb.plot_ohlc(ohlc_data=ohlc_data,
                            ohlc_key=ohlc_key)

    @staticmethod
    @validate_arguments
    def ohlc_table(ohlc_data: dict[str, pandas_DataFrame],
                   ohlc_key: int = 0) -> pd.DataFrame:
        """
        Creates a custom DataFrame to be displayed by Dash server.

        Args:
            ohlc_data (dict[key: DataFrame]): OHLC Data
            ohlc_key (int): Dictionary key

        Returns:
            DataFrame
        """
        return pb.ohlc_table(ohlc_data=ohlc_data,
                             ohlc_key=ohlc_key)

    @staticmethod
    @validate_arguments
    def plot_orderbook(orderbooks: dict[str, dict[str, pandas_DataFrame]],
                       orderbook_key: Union[int, str] = 0,
                       timestamp: Union[int, str] = -1) -> Union[go.Figure, None]:
        """
        Creates a plotly plot of a single order book to be displayed by browser or Dash server. Use order book_key
        to specify an order book and timestamp to specify the timeframe to display.

        Args:
            orderbooks (dict{key: DataFrame}): Order books
            orderbook_key (int): Dictionary key
            timestamp (int): Order book Key

        Returns:
            Plotly plot
        """
        return pb.plot_orderbook(orderbooks=orderbooks,
                                 orderbook_key=orderbook_key,
                                 timestamp=timestamp)

    @staticmethod
    @validate_arguments
    def plot_volume_history(trade_data: dict[str, pandas_DataFrame],
                            trade_key: Union[int, str] = 0) -> Union[go.Figure, None]:
        """
        Creates a plotly plot of the trade volume for a single contract to be displayed by browser or Dash server.

        Args:
            trade_data (dict{key: DataFrame}):  Trade Data
            trade_key (int/str): Dictionary key

        Returns:
            Plotly plot
        """
        return pb.plot_volume_history(trade_data=trade_data,
                                      trade_key=trade_key)


class ApiExporter(BaseExporter):
    """
    Exporter class for interaction with the PowerBot API.

    This class can/ should be used when:
        - the requested data is recent enough to still be stored in the PowerBot instance (see data retention policy)
        - the requested data is fairly small in size (multiple hours, not multiple day -> extensive loading time &
          constant strain on the API rate limit)
        - the requested data is already stored in the local __pb_cache__ and has been loaded via API.

    ATTENTION: if you try to load data from your cache and the original data is already purged from your instance,
    you are no longer able to create an index of contract IDs to load the local data with. Should this occur, please
    load the data in question via the HistoryExporter.
    """
    api_key: str = Field(description="A Standard API Key",
                         example="XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXX")
    host: str = Field(description="URL of the PowerBot instance to connect to",
                      example="https://backup.powerbot-trading.com/{COMPANY NAME}/{EXCHANGE}/v2/api")

    client: ApiClient = Field(description="Placeholder value, do not overwrite", default=None)

    def __init__(self, **data: Any):
        super().__init__(**data)

        # Init client
        self.client = pb.init_client(api_key=self.api_key, host=self.host)

    @root_validator
    def check_credentials(cls, values):
        api_key, host = values.get('api_key'), values.get('host')

        pattern_key = "\w{8}-\w{4}-\w{4}-\w{4}-\w{12}"
        pattern_host = "https://\w{4,7}.powerbot-trading.com/\w+/\w+/v\d{1}/api"

        if not re.match(pattern_key, api_key) or not re.match(pattern_host, host):
            raise ValueError("Your credentials do not conform to the necessary formats")

        return values

    @validate_arguments
    def get_contract_ids(self,
                         time_from: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss", default=None),
                         time_till: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss", default=None),
                         contract_ids: list[str] = None,
                         contract_time: str = "all",
                         products: list[str] = None,
                         allow_udc: bool = False,
                         delivery_areas: list[str] = None,
                         return_contract_objects: bool = False) -> Union[dict[str, list[str]], tuple[dict, dict]]:
        """
        Loads all contract IDs for specified timeframe, either by hours, single day or multiple days. Alternatively, a list
        of contract IDs can be passed instead of a timeframe, returning a dictionary of contract IDs compatible with all other
        functions of the Backtesting package. If return_contract_objects is True, a dictionary of contract objects will be
        returned as well.

        If time_from includes hh:mm:ss
            loads data between both times in timesteps according to contract_time

        If time_from doesn't include hh:mm:ss
            loads data between both dates in timesteps according to contract_time

        Args:
            time_from (datetime): yyyy-mm-dd hh:mm:ss
            time_till (datetime): yyyy-mm-dd hh:mm:ss
            contract_ids (list[str]): Optionally, a list of specific contract IDs can be passed to return contract objects
            contract_time (str): all, hourly, half-hourly or quarter-hourly (all includes UDC)
            products (list): Optional list of specific products to return
            allow_udc (bool): True if user-defined contracts (block products) should also be returned on contract_time = all
            delivery_areas (list): List of EIC-codes
            return_contract_objects (bool): If True, returns complete Contract object

        Returns:
            dict{key: (list[str])}: Dictionary of Contract IDs
            (Optional) Contract: Contract Object
        """
        return pb.get_contract_ids(api_client=self.client,
                                   time_from=time_from,
                                   time_till=time_till,
                                   contract_ids=contract_ids,
                                   contract_time=contract_time,
                                   products=products,
                                   allow_udc=allow_udc,
                                   delivery_areas=delivery_areas,
                                   return_contract_objects=return_contract_objects)

    @validate_arguments
    def get_public_trades(self,
                          contract_ids: dict[str, list[str]],
                          delivery_area: str,
                          contract_time: str,
                          iteration_delay: float = 0.4,
                          serialize_data: bool = True,
                          add_vwap: bool = False,
                          use_cached_data: bool = True,
                          caching: bool = True,
                          gzip_files: bool = True,
                          as_csv: bool = False) -> dict[str, pd.DataFrame]:
        """
        Load trade data for given contract IDs. If add_vwap is True, VWAP will be calculated for each trade, incorporating
        all previous trades.

        Args:
            contract_ids (dict): Dictionary of Contract IDs
            delivery_area (str): EIC Area Code for Delivery Area (not needed when loading from historic cache)
            contract_time (str): all, hourly, half-hourly or quarter-hourly
            iteration_delay (float): Optional delay between iterations to prevent hitting API rate limits
            serialize_data (bool): If False, request is received without serialization. Recommended for large data collections
            add_vwap (bool): If True, additional VWAP parameters will be added to each dataframe
            use_cached_data (bool): If True, function tries to load data from cache wherever possible
            caching (bool): True if data should be cached
            gzip_files (bool): True if cached files should be gzipped
            as_csv (bool): if True, will save files as CSV, additionally to JSON

        Returns:
            dict{key: DataFrame}: Dictionary of DataFrames
        """
        return pb.get_public_trades(api_client=self.client,
                                    contract_ids=contract_ids,
                                    contract_time=contract_time,
                                    delivery_area=delivery_area,
                                    iteration_delay=iteration_delay,
                                    serialize_data=serialize_data,
                                    add_vwap=add_vwap,
                                    use_cached_data=use_cached_data,
                                    caching=caching,
                                    gzip_files=gzip_files,
                                    as_csv=as_csv)

    @validate_arguments
    def get_public_trades_by_days(self,
                                  previous_days: int,
                                  delivery_area: str,
                                  time_from: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss", default=None),
                                  contract_time: str = None,
                                  contract_id: str = None) -> dict[str, pd.DataFrame]:
        """
        Gets the contract ID specified by a timeframe or directly by ID and load all trade data for this contract and all
        contracts in the same timeframe for X previous days.

        Args:
            time_from (datetime): YYYY-MM-DD hh:mm:ss
            previous_days (int): Amount of previous days to load data for
            delivery_area (str): EIC Area Code for Delivery Area
            contract_time (str): hourly, half-hourly or quarter-hourly
            contract_id (str): specific contract ID

        Returns:
            dict{key: DataFrame}: Dictionary of DataFrames
        """
        return pb.get_public_trades_by_days(api_client=self.client,
                                            previous_days=previous_days,
                                            delivery_area=delivery_area,
                                            time_from=time_from,
                                            contract_time=contract_time,
                                            contract_id=contract_id)

    @validate_arguments
    def get_contract_history(self,
                             contract_ids: dict[str, list[str]],
                             delivery_area: str,
                             iteration_delay: float = 0.4,
                             serialize_data: bool = True,
                             use_cached_data: bool = True,
                             caching: bool = True,
                             gzip_files: bool = True,
                             as_csv: bool = False) -> dict[str, pd.DataFrame]:
        """
        Load contract history for given contract IDs.

        Args:
            contract_ids (dict): Dictionary of Contract IDs
            delivery_area (str): EIC Area Code for Delivery Area (not needed when loading from historic cache)
            iteration_delay (float): Optional delay between iterations to prevent hitting API rate limits
            serialize_data (bool): If False, request is received without serialization. Recommended for large data collections
            use_cached_data (bool): If True, function tries to load data from cache wherever possible
            caching (bool): True if data should be cached
            gzip_files (bool): True if cached files should be gzipped
            as_csv (bool): if True, will save files as CSV, additionally to JSON

        Returns:
            dict{key: DataFrame}: Dictionary of DataFrames
        """
        return pb.get_contract_history(api_client=self.client,
                                       contract_ids=contract_ids,
                                       delivery_area=delivery_area,
                                       iteration_delay=iteration_delay,
                                       serialize_data=serialize_data,
                                       use_cached_data=use_cached_data,
                                       caching=caching,
                                       gzip_files=gzip_files,
                                       as_csv=as_csv)

    @validate_arguments
    def get_signals(self,
                    time_from: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                    time_till: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                    delivery_area: str = None,
                    portfolio_id: list[str] = None) -> list[Signal]:
        """
        Function gathers all Signals received by the API in the specified time period and gathers them in a list.

        Args:
            time_from (datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
            time_till (datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
            delivery_area (str): EIC Area Code for Delivery Area
            portfolio_id (str): List of all portfolios that signals should be loaded from

        Returns:
            list[Signal]
        """
        return pb.get_signals(api_client=self.client,
                              time_from=time_from,
                              time_till=time_till,
                              delivery_area=delivery_area,
                              portfolio_id=portfolio_id)

    @validate_arguments
    def get_own_trades(self,
                       time_from: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                       time_till: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                       delivery_area: str = None,
                       portfolio_id: list[str] = None) -> list[Trade]:
        """
        Function to collect all Own Trades for the defined time period, either specific to portfolio and/or delivery area
        or all portfolios and delivery areas used API key has access to.

        Args:
            time_from (datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
            time_till (datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
            delivery_area (str): EIC Area Code for Delivery Area
            portfolio_id (list[str]): List of specific portfolio IDs to load trades from. If left out, will load from all IDs.

        Returns:
            list[Trade]
        """
        return pb.get_own_trades(api_client=self.client,
                                 time_from=time_from,
                                 time_till=time_till,
                                 delivery_area=delivery_area,
                                 portfolio_id=portfolio_id)

    @validate_arguments
    def get_internal_trades(self,
                            time_from: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                            time_till: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                            delivery_area: str = None,
                            portfolio_id: list[str] = None) -> list[InternalTrade]:
        """
        Function to collect all Internal Trades for the defined time period, either specific to portfolio and/or delivery
        area or all portfolios and delivery areas used API key has access to.

        Args:
            time_from (datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
            time_till (datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
            delivery_area (str): EIC Area Code for Delivery Area
            portfolio_id (list[str]): List of specific portfolio IDs to load trades from. If left out, will load from all IDs.

        Returns:
            list[InternalTrade]
        """
        return pb.get_internal_trades(api_client=self.client,
                                      time_from=time_from,
                                      time_till=time_till,
                                      delivery_area=delivery_area,
                                      portfolio_id=portfolio_id)

    @validate_arguments
    def get_own_orders(self,
                       delivery_area: str = None,
                       portfolio_id: list[str] = None,
                       active_only: bool = False) -> list[OwnOrder]:
        """
        Function to collect all available Own Orders, either specific to portfolio and/or delivery area or all portfolios
        and delivery areas used API key has access to.

        Args:
            delivery_area (str): EIC Area Code for Delivery Area
            portfolio_id (list[str]): List of specific portfolio IDs to load trades from. If left out, will load from all IDs.
            active_only (bool):  True if only active orders should be loaded. If False, loads also hibernate and inactive.

        Returns:
            list[OwnOrder]
        """
        return pb.get_own_orders(api_client=self.client,
                                 delivery_area=delivery_area,
                                 portfolio_id=portfolio_id,
                                 active_only=active_only)

    @validate_arguments
    def calc_trade_vwap(self,
                        contract_time: str,
                        delivery_area: str,
                        trade_data: dict[str, pandas_DataFrame] = None,
                        time_from: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss", default=None),
                        previous_days: int = 10,
                        contract_id: str = None,
                        index: str = "ID3") -> pd.DataFrame:
        """
        Function gets trades for a certain contract for X previous days in the same delivery period and calculates their
        VWAP for ID3 or ID1 or all. Generates a new list of trades for these contracts.
        Can take either a time period or a specific contract ID to load data for.

        If previous days is 0, only the trades for the original time period/ contract will be loaded.

        Can also be called directly from get_public_trades with parameter 'add_vwap' to add VWAP to loaded trades.

        Args:
            contract_time (str): hourly, half-hourly or quarter-hourly
            delivery_area (str): Area Code for Delivery Area
            trade_data (dict[str, pd.DataFrame]: Dictionary of Dataframes containing Contract Trade Data
            time_from (datetime): yyyy-mm-dd hh:mm:ss
            previous_days (int): Amount of previous days to load data
            contract_id (str): ID of specific Contract
            index (str): all, ID3, ID1

        Returns:
            DataFrame: Trade Data with added calculated fields
        """
        return pb.calc_trade_vwap(api_client=self.client,
                                  contract_time=contract_time,
                                  delivery_area=delivery_area,
                                  trade_data=trade_data,
                                  time_from=time_from,
                                  previous_days=previous_days,
                                  contract_id=contract_id,
                                  index=index)


class HistoryExporter(BaseExporter):
    """
    Exporter class for interaction with the PowerBot History API and the subsequently created local __pb_cache__.

    This class can/ should be used when:
        - the requested data has is older than at least 2-3 days and has already been made available via History API
        - the requested data is already stored in the local __pb_cache__ and has been loaded via History API.

    ATTENTION: loading historic data from the History API will create a json file containing all contract information
    for the respective day. If this file should be deleted, the HistoryExporter can no longer create an index of
    contract IDs and therefore not load anything from the local cache.
    """
    exchange: str = Field(description="The exchange data should be loaded for")
    delivery_area: str = Field(description="EIC-code of the delivery area data should be loaded for")

    client: HistoryApiClient = Field(description="Placeholder value, do not overwrite", default=None)

    def __init__(self, **data: Any):
        super().__init__(**data)

        # Init client
        self.client = pb.init_historic_client(self.exchange, self.delivery_area)

    @root_validator
    def check_credentials(cls, values):
        exchange, delivery_area = values.get('exchange'), values.get('delivery_area')

        assert exchange in EXCHANGES, "Exchange is not in allowed exchanges"
        assert delivery_area in EIC_CODES[exchange], "Delivery area is not in allowed delivery areas for this exchange"

        return values

    @validate_arguments
    def get_historic_data(self,
                          api_key: str,
                          day_from: Union[str, datetime],
                          day_to: Union[str, datetime] = None,
                          delivery_areas: list[str] = None,
                          cache_path: Path = None,
                          extract_files: bool = False,
                          process_data: bool = False,
                          skip_on_error: bool = False,
                          keep_zip_files: bool = False) -> Union[list, dict]:
        """
        Function loads all public data for specified days in the specified delivery area. Output is a zipped directory
        containing all files in JSON format. Optionally, zip file can be extracted automatically and processed to be
        compatible with other functions in the powerbot_backtesting package.

        Args:
            api_key (str): Specific history instance API key
            day_from (str): Datetime/ String in format YYYY-MM-DD
            day_to (str): Datetime/ String in format YYYY-MM-DD
            delivery_areas (list): List of EIC Area Codes for Delivery Areas
            cache_path (Path): Optional path for caching files
            extract_files (bool): True if zipped files should be extracted automatically (Warning: immense size increase)
            process_data (bool): True if extracted files should be processed to resemble files loaded via API
            skip_on_error (bool): True if all dates that cannot possibly be loaded (e.g. due to lack of access rights) are
            skipped if the difference between day_from and day_to is at least 2 days
            keep_zip_files (bool): True if zip-files should be kept after download

        Returns:
            list of loaded file paths | dict
        """
        return pb.get_historic_data(api_key=api_key,
                                    exchange=self.client.exchange,
                                    delivery_areas=delivery_areas if delivery_areas else [self.client.delivery_area],
                                    day_from=day_from,
                                    day_to=day_to,
                                    cache_path=cache_path,
                                    extract_files=extract_files,
                                    process_data=process_data,
                                    skip_on_error=skip_on_error,
                                    keep_zip_files=keep_zip_files)

    @validate_arguments
    def get_contract_ids(self,
                         time_from: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                         time_till: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                         contract_time: Optional[str] = "all",
                         products: Optional[list[str]] = None,
                         allow_udc: Optional[bool] = False) -> dict[str, list[str]]:
        """
        Loads all contract IDs for specified timeframe, either by hours, single day or multiple days from the local cache.
        The cached data has to have been loaded via get_historic_data, since this function utilizes the contract file
        as an index.

        If time_from includes hh:mm:ss
            loads data between both times in timesteps according to contract_time

        If time_from doesn't include hh:mm:ss
            loads data between both dates in timesteps according to contract_time

        Args:
            time_from (datetime): yyyy-mm-dd hh:mm:ss
            time_till (datetime): yyyy-mm-dd hh:mm:ss
            contract_time (str): all, hourly, half-hourly or quarter-hourly (all includes UDC)
            products (list): Optional list of specific products to return
            allow_udc (bool): True if user-defined contracts (block products) should also be returned on contract_time = all

        Returns:
            dict{key: (list[str])}: Dictionary of Contract IDs
        """
        return pb.get_historic_contract_ids(client=self.client,
                                            time_from=time_from,
                                            time_till=time_till,
                                            contract_time=contract_time,
                                            products=products if products else [],
                                            allow_udc=allow_udc)

    @validate_arguments
    def get_public_trades(self,
                          contract_ids: dict[str, list[str]],
                          contract_time: str,
                          add_vwap: bool = False,
                          as_csv: bool = False) -> dict[str, pd.DataFrame]:
        """
        Load trade data for given contract IDs from local cache. If add_vwap is True, VWAP will be calculated for each
        trade, incorporating all previous trades.

        Args:
            contract_ids (dict): Dictionary of Contract IDs
            contract_time (str): all, hourly, half-hourly or quarter-hourly
            add_vwap (bool): If True, additional VWAP parameters will be added to each dataframe
            as_csv (bool): if True, will save files as CSV, additionally to JSON

        Returns:
            dict{key: DataFrame}: Dictionary of DataFrames
        """
        return pb.get_public_trades(api_client=self.client,
                                    contract_ids=contract_ids,
                                    contract_time=contract_time,
                                    add_vwap=add_vwap,
                                    caching=True if as_csv else False,
                                    as_csv=as_csv)

    @validate_arguments
    def get_contract_history(self,
                             contract_ids: dict[str, list[str]],
                             as_csv: bool = False) -> dict[str, pd.DataFrame]:
        """
        Load contract history for given contract IDs from the local cache.

        Args:
            contract_ids (dict): Dictionary of Contract IDs
            as_csv (bool): if True, will save files as CSV, additionally to JSON

        Returns:
            dict{key: DataFrame}: Dictionary of DataFrames
        """
        return pb.get_contract_history(api_client=self.client,
                                       contract_ids=contract_ids,
                                       as_csv=as_csv)


class SQLExporter(BaseExporter):
    """
    Exporter class for interaction with a SQL database containing PowerBot data.

    This class can/ should be used when:
        - a database containing the structure as defined per PowerBot SQLImporter exists and contains data.

    Instructions for passing arguments to exporter functions:
        - String
            If keyword argument is a string, it will be simply compared with '='. Optionally, a mathematical/ SQL
            operator (LIKE, <, <=, >, >=, <>) can be passed within the string. This operator will be used instead of '='.

            Example:
                best_bid='> 0.00' -> best_bid > 0.00 / as_of="> 2020-09-20 00:00:00" -> as_of > '2020-09-20 00:00:00'

        - Tuple
            If keyword argument is a tuple, it will be checked, if parameter is one of the elements of the tuple.

            Example:
                exchange=("Epex","NordPool") -> exchange IN ('Epex','NordPool')

        - List
            If keyword argument is a list, each element will be checked if it is in the parameter.

            Example:
                portfolio_ids=["TP1","TP2"] -> (portfolio_id LIKE '%TP1%' OR portfolio_id LIKE '%TP2%')

        - Dictionary
            If keyword argument is a dictionary, all values will be extracted and put into a tuple. Afterwards, the
            behaviour is the same as with tuples.

            Example:
                exchange={1:"Epex",2:"NordPool"} -> exchange IN ("Epex","NordPool")

        - Datetime
            If keyword argument is a datetime, parameter will be searched for the exact time of the datetime argument.
            This will in most cases not provide a satisfying result, therefore it is recommended to pass a datetime as
            a string with an operator in front.

            Example:
                as_of=datetime.datetime(2020, 9, 30, 10, 0, 0) -> as_of = '2020-09-30 10:00:00'
    """
    db_type: str = Field(description="Database type")
    user: str = Field(description="Database user")
    password: str = Field(description="Database password")
    host: str = Field(description="Database host address")
    database: str = Field(description="Database name")
    port: int = Field(description="Database port")

    logger: None = Field(description="Placeholder value, do not overwrite", default=None)
    engine: None = Field(description="Placeholder value, do not overwrite", default=None)

    SQL_ERRORS = (
        InterfaceError,
        OperationalError,
        ProgrammingError,
        TimeoutError,
    )

    def __init__(self, **data: Any):
        super().__init__(**data)

        # Logging Setup
        logging.basicConfig(format="PowerBot_SQL_Exporter %(asctime)s %(levelname)-8s %(message)s",
                            level=logging.INFO)
        self.logger = logging.getLogger()

        # Initialize Connection
        self.engine = self.__create_sql_engine()

    @validator("db_type")
    def validate_db_type(cls, value):
        allowed_db_types = ['mysql', 'mariadb', 'postgresql', 'oracle', 'mssql', 'amazon_redshift', 'apache_drill',
                            'apache_druid', 'apache_hive', 'apache_solr', 'cockroachdb', 'cratedb', 'exasolution',
                            'firebird', 'ibm_db2', 'monetdb', 'snowflake', 'teradata_vantage']

        assert value in allowed_db_types, f"Database {value} is not in allowed database"

        return value

    def __install_packages(self):
        """
        Tests if required packages for chosen SQL database type are available and installs them if necessary.
        """
        db_packages = {"mysql": ["mysql-connector-python"],
                       "mariadb": ["PyMySQL"],
                       "postgresql": ["psycopg2"],
                       "oracle": ["cx-Oracle"],
                       "mssql": ["pyodbc"],
                       "amazon_redshift": ["sqlalchemy-redshift", "psycopg2"],
                       "apache_drill": ["sqlalchemy-drill"],
                       "apache_druid": ["pydruid"],
                       "apache_hive": ["PyHive"],
                       "apache_solr": ["sqlalchemy-solr"],
                       "cockroachdb": ["sqlalchemy-cockroachdb", "psycopg2"],
                       "cratedb": ["crate-python"],
                       "exasolution": ["sqlalchemy_exasol", "pyodbc"],
                       "firebird": ["sqlalchemy-firebird"],
                       "ibm_db2": ["ibm_db_sa"],
                       "monetdb": ["sqlalchemy_monetdb"],
                       "snowflake": ["snowflake-sqlalchemy"],
                       "teradata_vantage": ["teradatasqlalchemy"]}

        self.logger.info("Now installing the following necessary package(s):\n"
                        f"{db_packages[self.db_type]}")

        import subprocess
        import sys
        for pkg in db_packages[self.db_type]:
            subprocess.check_call([sys.executable, "-m", "pip", "install", pkg])

        return self.__create_sql_engine()

    def __create_sql_engine(self):
        """
        Initializes connection to SQL database
        """
        db_types = {"mysql": "mysql+mysqlconnector",
                    "mariadb": "mariadb+pymysql",
                    "postgresql": "postgresql+psycopg2",
                    "oracle": "oracle+cx_oracle",
                    "mssql": "mssql+pyodbc",
                    "amazon_redshift": "redshift+psycopg2",
                    "apache_drill": "drill+sadrill",
                    "apache_druid": "druid",
                    "apache_hive": "hive",
                    "apache_solr": "solr",
                    "cockroachdb": "cockroachdb",
                    "cratedb": "crate",
                    "exasolution": "exa+pyodbc",
                    "firebird": "firebird",
                    "ibm_db2": "db2+ibm_db",
                    "monetdb": "monetdb",
                    "snowflake": "snowflake",
                    "teradata_vantage": "teradatasql"}

        if self.port:
            try:
                engine = create_engine(
                    f'{db_types[self.db_type]}://{self.user}:{self.password}@{self.host}:{self.port}/{self.database}')
            except:
                pass
        try:
            engine = create_engine(
                f'{db_types[self.db_type]}://{self.user}:{self.password}@{self.host}/{self.database}')

            # Test connection
            engine.connect()

        except (NoSuchModuleError, ModuleNotFoundError):
            self.logger.info("You currently do not have all the necessary packages installed to access a database of"
                            f" type {self.db_type}.")
            return self.__install_packages()

        except ProgrammingError:
            self.logger.error("Could not establish connection to database. Please recheck your credentials!")

        except InterfaceError:
            self.logger.error("Database is not available at the moment!")

        except Exception as e:
            raise SQLExporterError(f"Could not establish connection to database. Reason: \n{e}")

        self.logger.info(f"Connection to database '{self.database}' with user '{self.user}' established")
        self.logger.info("Connection ready to export data")

        return engine

    @contextmanager
    def __get_session(self):
        """
        Context manager to handle sessions connecting to the database.
        """
        try:
            session = Session(bind=self.engine)
        except self.SQL_ERRORS:
            session = Session(bind=self.__create_sql_engine())
        try:
            yield session
        finally:
            session.close()

    @validate_arguments
    def get_contracts(self,
                      as_dataframe: bool = True,
                      **kwargs) -> pd.DataFrame:
        """
        Exports contracts from SQL database. To use different mathematical/SQL operators, pass keyworded arguments
        as strings and include the desired operator followed by a space (e.g. revisions='<> 0').

        Following operators can be passed:
        LIKE, <, <=, >, >=, <>

        Following parameters can be passed as optional keyworded arguments:
        exchange, contract_id, product, type, undrlng_contracts, name, delivery_start, delivery_end, delivery_areas,
        predefined, duration, delivery_units

        Args:
            as_dataframe (bool): If False -> returns list
            **kwargs: any additional fields of SQL table

        Returns:
            list/ DataFrame: SQL query
        """
        allowed_kwargs = ["name", "delivery_areas", "delivery_start", "delivery_end", "delivery_areas", "type",
                          "predefined", "duration", "delivery_units", "contract_id", "exchange", "product",
                          "undrlng_contracts"]

        sql_params = self.__handle_sql_args(kwargs, allowed_kwargs)

        with self.__get_session() as session:
            result = session.execute(f"SELECT * FROM contracts{sql_params}").fetchall()

        if as_dataframe:
            output = pd.DataFrame(result)
            output = output.rename(
                columns={0: 'exchange', 1: 'contract_id', 2: 'product', 3: 'type', 4: 'undrlng_contracts',
                         5: 'name', 6: 'delivery_start', 7: 'delivery_end', 8: 'delivery_areas',
                         9: 'predefined', 10: 'duration', 11: 'delivery_units', 12: 'details'})
            return output

        return result

    @validate_arguments
    def get_contract_ids(self,
                         time_from: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                         time_till: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                         delivery_area: str = Field(description="EIC-code of delivery area"),
                         contract_time: str = "all",
                         exchange: str = "epex",
                         as_list: bool = False) -> dict[str, list[str]]:
        """
            Returns dictionary of contract IDs in a format compatible with backtesting pipeline.

        Args:
            time_from (datetime): yyyy-mm-dd hh:mm:ss
            time_till (datetime): yyyy-mm-dd hh:mm:ss
            delivery_area (str): EIC-Code
            contract_time (str): all, hourly, half-hourly or quarter-hourly
            exchange (str): Name of exchange in lowercase
            as_list (bool): True if output should be list of contract IDs

        Returns:
            dict{key: (list[str])}: Dictionary of Contract IDs
        """
        if not isinstance(time_from, datetime) or not isinstance(time_till, datetime):
            raise SQLExporterError("Please use datetime format for time_from & time_till!")

        with self.__get_session() as session:
            result = session.execute(f"SELECT delivery_start, delivery_end, contract_id FROM contracts "
                                     f"WHERE delivery_start >= '{time_from}' "
                                     f"AND delivery_end <= '{time_till}' "
                                     f"AND delivery_areas LIKE '%{delivery_area}%' "
                                     f"AND product IN {PRODUCTS[contract_time]} "
                                     f"AND exchange = '{exchange}'").fetchall()
            if not result:
                result = session.execute(f"SELECT delivery_start, delivery_end, contract_id FROM contracts "
                                         f"WHERE delivery_start >= '{time_from}' "
                                         f"AND delivery_end <= '{time_till}' "
                                         f"AND product IN {PRODUCTS[contract_time]} "
                                         f"AND exchange = '{exchange}'").fetchall()

        if not as_list:
            contract_ids = {f"{i[0].strftime(DATE_YMD_TIME_HM)} - {i[1].strftime(TIME_HM)}": [] for i in result}
            for i in result:
                contract_ids[f"{i[0].strftime(DATE_YMD_TIME_HM)} - {i[1].strftime(TIME_HM)}"].append(i[2])

            # Quality Check
            if not all(i for i in contract_ids.values()):
                raise SQLExporterError("There is no contract data for the specified timeframe!")
        else:
            contract_ids = [i for i in result]

        self.logger.info("Successfully exported contract ids")
        return contract_ids

    @validate_arguments
    def get_public_trades(self,
                          as_dataframe: bool = True,
                          delivery_area: list[str] = None,
                          **kwargs) -> Union[pd.DataFrame, dict[str, pd.DataFrame]]:
        """
        Exports trades from SQL database. To use different mathematical/SQL operators, pass keyworded arguments
        as strings and include the desired operator followed by a space (e.g. price='<> 0.00').

        Following operators can be passed:
        LIKE, <, <=, >, >=, <>

        Following parameters can be passed as optional keyworded arguments (kwargs):
        price, quantity, prc_x_qty, exchange, contract_id, trade_id, exec_time, api_timestamp, self_trade

        Args:
            as_dataframe (bool): If False -> returns list
            delivery_area (tuple[str]): Multiple delivery areas inside a tuple. Single del. area can be passed as a string
            **kwargs: any additional fields of SQL table

        Returns:
            list/ DataFrame: SQL query
        """
        allowed_kwargs = ["price", "quantity", "prc_x_qty", "exchange", "contract_id", "trade_id", "exec_time",
                          "api_timestamp", "self_trade"]

        sql_params = self.__handle_sql_args(kwargs, allowed_kwargs)
        if delivery_area:
            for i in delivery_area:
                sql_params += f" AND (buy_delivery_area = '{i}' OR sell_delivery_area = '{i}')"

        with self.__get_session() as session:
            result = session.execute(f"SELECT * FROM public_trades{sql_params}").fetchall()

        if as_dataframe:
            output = pd.DataFrame(result)
            if not output.empty:
                output = output.rename(columns={0: 'exchange', 1: 'contract_id', 2: 'trade_id', 3: 'api_timestamp',
                                                4: 'exec_time', 5: 'buy_delivery_area', 6: 'sell_delivery_area',
                                                7: 'price', 8: 'quantity', 9: 'prc_x_qty', 10: "currency",
                                                11: 'self_trade'})
                if not delivery_area:
                    raise SQLExporterError("Please specify at least one delivery area!")

                output['api_timestamp'] = pd.to_datetime(output['api_timestamp'], utc=True)
                output['exec_time'] = pd.to_datetime(output['exec_time'], utc=True)

                self.logger.info("Successfully exported trades")
                return self.__convert_dataframe("trades", output)
            else:
                raise SQLExporterError("There is no trade data for the specified timeframe!")

        self.logger.info("Successfully exported trades")

        return result

    @validate_arguments
    def get_contract_history(self,
                             as_dataframe: bool = True,
                             **kwargs) -> Union[pd.DataFrame, dict[str, pd.DataFrame]]:
        """
        Exports contract revisions from SQL database. To use different mathematical/SQL operators, pass keyworded arguments
        as strings and include the desired operator followed by a space (e.g. best_bid='<> 0.00').

        Following operators can be passed:
        LIKE, <, <=, >, >=, <>

        Following parameters can be passed as optional keyworded arguments:
        exchange, contract_id, exchange, delivery_area, revision_no, as_of, best_bid, best_bid_qty, best_ask,
        best_ask_qty, vwap, high, low, last_price, last_qty, last_trade_time, volume, delta

        Args:
            as_dataframe (bool): If False -> returns list
            **kwargs: any additional fields of SQL table

        Returns:
            list/ DataFrame: SQL query
        """
        allowed_kwargs = ["exchange", "contract_id", "delivery_area", "revision_no", "as_of", "best_bid",
                          "best_bid_qty", "best_ask", "best_ask_qty", "vwap", "high", "low", "last_price",
                          "last_qty", "last_trade_time", "volume", "delta"]

        sql_params = self.__handle_sql_args(kwargs, allowed_kwargs)

        with self.__get_session() as session:
            result = session.execute(f"SELECT * FROM contract_revisions{sql_params}").fetchall()

        if as_dataframe:
            output = pd.DataFrame(result)
            if not output.empty:
                output = output.rename(columns={0: 'exchange', 1: 'contract_id', 2: 'delivery_area', 3: 'revision_no',
                                                4: 'as_of', 5: 'best_bid', 6: 'best_bid_qty', 7: 'best_ask',
                                                8: 'best_ask_qty', 9: 'vwap', 10: 'high', 11: 'low', 12: 'last_price',
                                                13: 'last_qty', 14: "last_trade_time", 15: 'volume', 16: 'delta',
                                                17: 'bids', 18: 'asks'})
                if "delivery_area" not in kwargs:
                    raise SQLExporterError("Please specify one specific delivery area!")
                self.logger.info("Successfully exported contract history")
                return self.__convert_dataframe("orders", output)
            else:
                raise SQLExporterError("There is no order data for the specified timeframe!")

        self.logger.info("Successfully exported contract history")

        return result

    @validate_arguments
    def get_own_trades(self,
                       delivery_area: list[str] = None,
                       as_dataframe: bool = True,
                       as_objects: bool = False,
                       **kwargs) -> Union[pd.DataFrame, list[Trade]]:
        """
        Exports Own Trades from SQL database. To use different mathematical/SQL operators, pass keyworded arguments
        as strings and include the desired operator followed by a space (e.g. position_short='<> 0.00').

        Following operators can be passed:
        LIKE, <, <=, >, >=, <>

        Following parameters can be passed as optional keyworded arguments (kwargs):
        exchange, contract_id, contract_name, prod, delivery_start, delivery_end, trade_id, api_timestamp, exec_time,
        buy, sell, price, quantity, state, buy_delivery_area, sell_delivery_area, buy_order_id, buy_clOrderId, buy_txt,
        buy_user_code, buy_member_id, buy_aggressor_indicator, buy_portfolio_id, sell_order_id, sell_clOrderId,
        sell_txt, sell_user_code, sell_member_id, sell_aggressor_indicator, sell_portfolio_id, self_trade, pre_arranged,
        pre_arrange_type

        Args:
            delivery_area (tuple[str]): Multiple delivery areas inside a tuple. Single del. area can be passed as a string
            as_dataframe (bool): True if output should be DataFrame
            as_objects (bool): True if output should be list of OwnTrades
            **kwargs: any additional fields of SQL table

        Returns:
            list: SQL query
        """

        allowed_kwargs = ["exchange", "contract_id", "contract_name", "prod", "delivery_start", "delivery_end",
                          "trade_id",
                          "api_timestamp", "exec_time", "buy", "sell", "price", "quantity", "state",
                          "buy_delivery_area",
                          "sell_delivery_area", "buy_order_id", "buy_clOrderId", "buy_txt", "buy_user_code",
                          "buy_member_id",
                          "buy_aggressor_indicator", "buy_portfolio_id", "sell_order_id", "sell_clOrderId", "sell_txt",
                          "sell_user_code", "sell_member_id", "sell_aggressor_indicator", "sell_portfolio_id",
                          "self_trade",
                          "pre_arranged", "pre_arrange_type"]

        sql_params = self.__handle_sql_args(kwargs, allowed_kwargs)

        if delivery_area:
            for i in delivery_area:
                sql_params += f" AND (buy_delivery_area = '{i}' OR sell_delivery_area = '{i}')"

        with self.__get_session() as session:
            result = session.execute(f"SELECT * FROM own_trades{sql_params}").fetchall()

        self.logger.info("Successfully exported own trades")

        # Convert data back to Trade objects
        if result and as_objects:
            own_trades = [Trade(exchange=i[0],
                                contract_id=i[1],
                                contract_name=i[2],
                                prod=i[3],
                                delivery_start=i[4],
                                delivery_end=i[5],
                                trade_id=i[6],
                                api_timestamp=i[7],
                                exec_time=i[8],
                                buy=i[9],
                                sell=i[10],
                                price=i[11],
                                quantity=i[12],
                                delivery_area=i[13],
                                state=i[14],
                                buy_delivery_area=i[15],
                                sell_delivery_area=i[16],
                                buy_order_id=i[17],
                                buy_cl_order_id=i[18],
                                buy_txt=i[19],
                                buy_user_code=i[20],
                                buy_member_id=i[21],
                                buy_aggressor_indicator=i[22],
                                buy_portfolio_id=i[23],
                                sell_order_id=i[24],
                                sell_cl_order_id=i[25],
                                sell_txt=i[26],
                                sell_user_code=i[27],
                                sell_member_id=i[28],
                                sell_aggressor_indicator=i[29],
                                sell_portfolio_id=i[30],
                                self_trade=i[31],
                                pre_arranged=i[32],
                                pre_arrange_type=i[33])
                          for i in result]
            return own_trades

        if result and as_dataframe:
            return pd.DataFrame(result)

        return result

    @validate_arguments
    def get_internal_trades(self,
                            delivery_area: tuple[str] = None,
                            as_dataframe: bool = True,
                            as_objects: bool = False,
                            **kwargs) -> Union[pd.DataFrame, list[InternalTrade]]:
        """
        Exports Internal Trades from SQL database. To use different mathematical/SQL operators, pass keyworded arguments
        as strings and include the desired operator followed by a space (e.g. position_short='<> 0.00').

        Following operators can be passed:
        LIKE, <, <=, >, >=, <>

        Following parameters can be passed as optional keyworded arguments (kwargs):
        exchange, contract_id, contract_name, prod, delivery_start, delivery_end, internal_trade_id, api_timestamp,
        exec_time, price, quantity, state, buy_delivery_area, sell_delivery_area, buy_order_id, buy_clOrderId, buy_txt,
        buy_aggressor_indicator, buy_portfolio_id, sell_order_id, sell_clOrderId, sell_txt, sell_aggressor_indicator,
        sell_portfolio_id

        Args:
            delivery_area (tuple[str]): Multiple delivery areas inside a tuple. Single del. area can be passed as a string
            as_dataframe (bool): True if output should be DataFrame
            as_objects (bool): True if output should be list of InternalTrades
            **kwargs: any additional fields of SQL table

        Returns:
            list: SQL query
        """
        allowed_kwargs = ["exchange", "contract_id", "contract_name", "prod", "delivery_start", "delivery_end",
                          "internal_trade_id", "api_timestamp", "exec_time", "price", "quantity", "state",
                          "buy_delivery_area", "sell_delivery_area", "buy_order_id", "buy_clOrderId", "buy_txt",
                          "buy_aggressor_indicator", "buy_portfolio_id", "sell_order_id", "sell_clOrderId", "sell_txt",
                          "sell_aggressor_indicator", "sell_portfolio_id"]

        sql_params = self.__handle_sql_args(kwargs, allowed_kwargs)

        if delivery_area:
            for i in delivery_area:
                sql_params += f" AND (buy_delivery_area = '{i}' OR sell_delivery_area = '{i}')"

        with self.__get_session() as session:
            result = session.execute(f"SELECT * FROM internal_trades{sql_params}").fetchall()

        self.logger.info("Successfully exported internal trades")

        # Convert data back to InternalTrade objects
        if result and as_objects:
            internal_trades = [InternalTrade(exchange=i[0],
                                             contract_id=i[1],
                                             contract_name=i[2],
                                             prod=i[3],
                                             delivery_start=i[4],
                                             delivery_end=i[5],
                                             internal_trade_id=i[6],
                                             api_timestamp=i[7],
                                             exec_time=i[8],
                                             price=i[9],
                                             quantity=i[10],
                                             buy_delivery_area=i[11],
                                             sell_delivery_area=i[12],
                                             buy_order_id=i[13],
                                             buy_cl_order_id=i[14],
                                             buy_txt=i[15],
                                             buy_aggressor_indicator=i[16],
                                             buy_portfolio_id=i[17],
                                             sell_order_id=i[18],
                                             sell_cl_order_id=i[19],
                                             sell_txt=i[20],
                                             sell_aggressor_indicator=i[21],
                                             sell_portfolio_id=i[22])
                               for i in result]
            return internal_trades

        if result and as_dataframe:
            return pd.DataFrame(result)

        return result

    @validate_arguments
    def get_signals(self,
                    time_from: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                    time_till: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                    as_dataframe: bool = True,
                    as_objects: bool = False,
                    **kwargs) -> Union[pd.DataFrame, list[Signal]]:
        """
        Exports signals from SQL database. To use different mathematical/SQL operators, pass keyworded arguments
        as strings and include the desired operator followed by a space (e.g. position_short='<> 0.00').

        Following operators can be passed:
        LIKE, <, <=, >, >=, <>

        Following parameters can be passed as optional keyworded arguments:
        id, source, received_at, revision, delivery_areas, portfolio_ids, tenant_id, position_short,
        position_long, value

        Args:
            time_from (datetime): yyyy-mm-dd hh:mm:ss
            time_till (datetime): yyyy-mm-dd hh:mm:ss
            as_dataframe (bool): True if output should be DataFrame
            as_objects (bool): True if output should be list of Signals
            **kwargs: any additional fields of SQL table

        Returns:
            list: SQL query
        """
        allowed_kwargs = ["id", "source", "received_at", "revision", "delivery_areas", "portfolio_ids", "tenant_id",
                          "position_short", "position_long", "value"]

        sql_params = self.__handle_sql_args(kwargs, allowed_kwargs)
        sql_op = "AND" if sql_params else "WHERE"

        with self.__get_session() as session:
            result = session.execute(f"SELECT * FROM signals{sql_params} "
                                     f"{sql_op} delivery_start >= '{time_from}' "
                                     f"AND delivery_end <= '{time_till}'").fetchall()

        self.logger.info("Successfully exported signals")

        # Convert data back to InternalTrade objects
        if result and as_objects:
            signals = [Signal(id=i[0],
                              source=i[1],
                              received_at=i[2],
                              revision=i[3],
                              delivery_start=i[4],
                              delivery_end=i[5],
                              portfolio_ids=i[6],
                              tenant_id=i[7],
                              position_short=i[8],
                              position_long=i[9],
                              value=i[10])
                       for i in result]

            return signals

        if result and as_dataframe:
            return pd.DataFrame(result)

        return result

    @validate_arguments
    def send_raw_sql(self,
                     sql_statement: str):
        """
        Function allows for raw SQL queries to be sent to the database.

        Args:
            sql_statement (str): SQL query

        Returns:

        """
        with self.__get_session() as session:
            try:
                result = session.execute(sql_statement).fetchall()
            except self.SQL_ERRORS as e:
                return self.logger.error(e)
        return result

    @staticmethod
    def __handle_sql_args(kwargs,
                          allowed_kwargs: list[str]) -> str:
        """
        Handles incoming arguments by adjusting them to be compatible with SQL.

        Args:
            kwargs: **kwargs of export functions
            allowed_kwargs (list[str]): list of allowed kwargs

        Returns:
            str: SQL request
        """
        if not all(arg for arg in kwargs.values()):
            raise SQLExporterError("Some of your input values are invalid or empty!")
        sql_params = ""
        operators = ["LIKE", "<", "<=", ">", ">=", "<>"]

        for keyword, argument in kwargs.items():
            op = "="
            sql_statement = "WHERE" if sql_params == "" else "AND"

            if keyword not in allowed_kwargs:
                raise SQLExporterError(f"{keyword} not in allowed keywords. Allowed keywords: {allowed_kwargs}")
            else:
                if isinstance(argument, str):
                    # Check For SQL Commands Or Mathematical Operators
                    if any(x in argument for x in operators):
                        if len(argument.split(" ")) > 2:
                            op = argument.split(" ")[0]
                            argument = argument.replace(f"{op} ", "")
                        else:
                            op, argument = argument.split(" ")
                        if op == "LIKE":
                            argument = f"%{argument}%"
                        try:
                            datetime.strptime(argument, DATE_YMD_TIME_HMS)
                        except:
                            pass
                elif isinstance(argument, tuple):
                    if len(argument) == 1:
                        argument = argument[0]
                    else:
                        op = "IN"
                elif isinstance(argument, list):
                    for nr, element in enumerate(argument):
                        if not nr:
                            if element == argument[-1]:
                                sql_params += f" {sql_statement} ({keyword} LIKE '%{element}%')"
                            else:
                                sql_params += f" {sql_statement} ({keyword} LIKE '%{element}%'"
                        elif element == argument[-1]:
                            sql_params += f" OR {keyword} LIKE '%{element}%')"
                        else:
                            sql_params += f" OR {keyword} LIKE '%{element}%'"
                    continue

                elif isinstance(argument, dict):
                    op = "IN"
                    temp_list = []
                    for value in argument.values():
                        for item in value:
                            temp_list.append(item)
                    argument = tuple(temp_list)
                try:
                    if not isinstance(argument, tuple) and keyword != "contract_id":
                        argument = float(argument)
                    sql_params += f" {sql_statement} {keyword} {op} {argument}"
                except:
                    sql_params += f" {sql_statement} {keyword} {op} '{argument}'"

        return sql_params

    def __convert_dataframe(self,
                            df_type: str,
                            dataframe: pandas_DataFrame) -> dict[str, pd.DataFrame]:
        """
        Function to convert dataframe to required format to be processed by backtesting data pipeline.

        Args:
            df_type (str): orders/trades/orderbooks
            dataframe (DataFrame): DataFrame containing exported Data

        Returns:
            dict{key: DataFrame}: Dictionary of DataFrames
        """
        output = {}

        contract_ids = dataframe.contract_id.unique().tolist()
        contracts = self.get_contracts(contract_id=contract_ids)

        if df_type == "trades":
            dataframe = dataframe.astype({'price': 'float', 'quantity': 'float'})

        elif df_type == "orders":
            dataframe["bids"] = [json.loads(i) if i else None for i in dataframe.bids.tolist()]
            dataframe["asks"] = [json.loads(i) if i else None for i in dataframe.asks.tolist()]

        # 	dataframe.drop(columns=["bids", "asks"])
        # 	orders = []
        # 	for nr, val in enumerate(bids):
        # 		orders.append({"bid": val, "ask": asks[nr]})
        # 	dataframe["orders"] = orders

        for row_nr, row_id in enumerate(contracts.contract_id):
            key = f"{contracts.iloc[row_nr].delivery_start.strftime(DATE_YMD_TIME_HM)} - " \
                  f"{contracts.iloc[row_nr].delivery_end.strftime(TIME_HM)}"

            if key not in [*output]:
                output[key] = dataframe[dataframe["contract_id"] == row_id]
            else:
                output[key].append(dataframe[dataframe["contract_id"] == row_id])

        return output



import gzip
import hashlib
import json
from collections import defaultdict
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Union
from zipfile import ZipFile, BadZipFile

import pandas as pd
import requests
from powerbot_client.exceptions import ApiException

from powerbot_backtesting.exceptions import ChecksumError, NotInCacheError
from powerbot_backtesting.historic_data_processing import process_historic_data
from powerbot_backtesting.models import HistoryApiClient
from powerbot_backtesting.utils import _find_cache
from powerbot_backtesting.utils.constants import *


def get_historic_data(api_key: str,
                      exchange: str,
                      delivery_areas: list[str],
                      day_from: Union[str, datetime],
                      day_to: Union[str, datetime] = None,
                      cache_path: Path = None,
                      extract_files: bool = False,
                      process_data: bool = False,
                      skip_on_error: bool = False,
                      keep_zip_files: bool = False) -> Union[list, dict]:
	"""
	Function loads all public data for specified days in the specified delivery area. Output is a zipped directory
	containing all files in JSON format. Optionally, zip file can be extracted automatically and processed to be
	compatible with other functions in the powerbot_backtesting package.

	Args:
		api_key (str): Specific history instance API key
		exchange (str): One of the following: epex, hupx, tge, nordpool, southpool, ibex
		delivery_areas (list): List of EIC Area Codes for Delivery Areas
		day_from (str): Datetime/ String in format YYYY-MM-DD
		day_to (str): Datetime/ String in format YYYY-MM-DD
		cache_path (Path): Optional path for caching files
		extract_files (bool): True if zipped files should be extracted automatically (Warning: immense size increase)
		process_data (bool): True if extracted files should be processed to resemble files loaded via API
		skip_on_error (bool): True if all dates that cannot possibly be loaded (e.g. due to lack of access rights) are
		skipped if the difference between day_from and day_to is at least 2 days
		keep_zip_files (bool): True if zip-files should be kept after download

	Returns:
		list of loaded file paths | dict
	"""
	# Validity check
	if not isinstance(day_from, datetime):
		try:
			day_from = datetime.strptime(day_from, DATE_YMD)
		except ValueError:
			raise ValueError("day_from needs to be a date or a string in YYYY-MM-DD format")
	if day_to and not isinstance(day_to, datetime):
		try:
			day_to = datetime.strptime(day_to, DATE_YMD)
		except ValueError:
			raise ValueError("day_to needs to be a date or a string in YYYY-MM-DD format")

	delivery_areas = delivery_areas if isinstance(delivery_areas, list) else [delivery_areas]
	cache_path = _find_cache() if not cache_path else cache_path
	cache_path = Path(cache_path) if not isinstance(cache_path, Path) else cache_path
	headers = {"accept": "application/zip", "X-API-KEY": api_key}
	day_to = day_to if day_to else day_from
	skip_on_error = True if skip_on_error and day_to and day_to - day_from >= timedelta(days=2) else False

	zipfiles = []
	extracted_files = {}
	retry = 0

	while day_from <= day_to:
		# While False, days will continue with iteration
		prevent_update = False

		for del_area in delivery_areas:
			host = f"https://history.powerbot-trading.com/history/{exchange}/{del_area}/{day_from.strftime(DATE_YMD)}"

			# Filepath
			filepath = cache_path.joinpath(f"history/{exchange}_{del_area}/{day_from.strftime(DATE_YM)}")

			# File
			filename = f"{day_from.strftime(DATE_MD)}_public_data.zip"
			zipfiles.append(f"{del_area}_{filename.strip('.zip')}")

			# Skip if file exists
			if not filepath.joinpath(filename).exists() and not filepath.joinpath(day_from.strftime(DATE_MD)).exists():
				# Load file
				r = requests.get(host, headers=headers, stream=True)
				m = hashlib.sha256()

				if skip_on_error and r.status_code in [204, 403, 404]:
					continue
				if r.status_code == 503:
					raise ApiException(status=503, reason="Service unavailable or API rate limit exceeded")
				if r.status_code == 404:
					raise ApiException(status=404,
					                   reason=f"Data for '{day_from}' in {del_area} has not been exported yet")
				if r.status_code == 403:
					raise ApiException(status=403, reason="Currently used API Key does not have access to this data")
				if r.status_code == 204:
					raise ApiException(status=204, reason=f"There is no data for '{day_from}' in {del_area}")

				# Create filepath only if file is valid
				filepath.mkdir(parents=True, exist_ok=True)

				with open(filepath.joinpath(filename), 'wb') as fd:
					for chunk in r.iter_content(chunk_size=128):
						m.update(chunk)
						fd.write(chunk)

				expected_hash = \
					json.loads(
						[i for i in requests.get(host + "/sha256", headers=headers, stream=True).iter_lines()][0])[
						"sha_256"]

				if not expected_hash == m.hexdigest():
					if retry < 3:  # Retry 3 times
						filepath.joinpath(filename).unlink(missing_ok=False)
						retry += 1
						prevent_update = True
						continue
					if skip_on_error:  # Skip
						filepath.joinpath(filename).unlink(missing_ok=False)
					else:
						filepath.joinpath(filename).unlink(missing_ok=False)
						raise ChecksumError(
							"Corrupted file: expected sha256 checksum does not match sha256 of received files. "
							"Please try again.")

			# Extraction
			if extract_files and not filepath.joinpath(day_from.strftime(DATE_MD)).exists():
				try:
					with ZipFile(filepath.joinpath(filename), 'r') as _zip:
						_zip.extractall(filepath.joinpath(day_from.strftime(DATE_MD)))
					if not keep_zip_files:
						# Delete Zip
						filepath.joinpath(filename).unlink()

				except BadZipFile:
					raise TypeError(f"The created file for day '{day_from}' is faulty. Please try to load it again")

			if extract_files:
				# Add to dictionary
				extracted_files[f"{del_area}_{filename.strip('.zip')}"] = [str(e) for e in filepath.joinpath(
					day_from.strftime(DATE_MD)).iterdir() if e.is_file()]

			retry = 0
		day_from = day_from + timedelta(days=1) if not prevent_update else day_from

	if not zipfiles:
		return []
	if extract_files:
		if process_data:
			return process_historic_data(extracted_files, exchange, keep_zip_files)
		return extracted_files
	return zipfiles


def get_history_key_info(api_key: str) -> dict:
	"""
	Returns information for the specified History API Key

	Args:
		api_key (str): History API Key

	Returns:
		dict
	"""
	headers = {"accept": "application/json", "X-API-KEY": api_key}
	host = "https://history.powerbot-trading.com/api-key"

	r = requests.get(host, headers=headers, stream=True)

	if r.status_code == 503:
		raise ApiException(status=503, reason="Service unavailable or API rate limit exceeded")
	if r.status_code == 400:
		raise ApiException(status=400, reason="Request could not be processed, please check your API Key")
	return r.json()


def init_historic_client(exchange: str,
                         delivery_area: str) -> HistoryApiClient:
	"""
	Initializes historic client containing information to filter for cached historic data. When using a HistoryApiClient
	instead of an ApiClient, all other function will react accordingly and only load data from the local __pb_cache__.

	Args:
		exchange (str): One of the following: epex, hupx, tge, nordpool, southpool, ibex
		delivery_area (str):  EIC Area Code for Delivery Area

	Returns:
		HistoryApiClient
	"""
	return HistoryApiClient(exchange=exchange, delivery_area=delivery_area)


def get_historic_contract_ids(client: HistoryApiClient,
                              time_from: datetime,
                              time_till: datetime,
                              contract_time: str,
                              products: list[str],
                              allow_udc: bool):
	"""
	Function mimics get_contract_ids but loads contract IDs from a cached index file instead of the API. This is
	necessary when trying to load data that is older than the data retention policy for production instances.

	Args:
		client: HistoryApiClient
		time_from (datetime): yyyy-mm-dd hh:mm:ss
		time_till (datetime): yyyy-mm-dd hh:mm:ss
		contract_time (str): all, hourly, half-hourly or quarter-hourly (all includes UDC)
		products (list): List of specific products to return
		allow_udc (bool): True if user-defined contracts (block products) should also be returned on contract_time = all

	Returns:
		dict{key: (list[str])}: Dictionary of Contract IDs
	"""
	# Setup
	timeframes = {"all": 15, "hourly": 60, "half-hourly": 30, "quarter-hourly": 15}
	ct = timeframes[contract_time]
	products = PRODUCTS[contract_time] if not products else products
	contracts = defaultdict(list)
	time_range = []

	# Split time up if range covers multiple days
	while time_from.date() < time_till.date():
		time_range.append((time_from, time_from.replace(hour=0, minute=0, second=0) + timedelta(days=1)))
		time_from = time_from.replace(hour=0, minute=0, second=0) + timedelta(days=1)
	else:
		if not time_from == time_till:
			time_range.append((time_from, time_till))

	# Find __cache__ directory
	cache_path = _find_cache()

	for tr in time_range:
		# Extract date information
		year_month = tr[0].strftime(DATE_YM)
		day_month = tr[0].strftime(DATE_MD)

		# Construct path to index file
		file_path = cache_path.joinpath(f"prod\{client.exchange}_{client.delivery_area}\{year_month}\{day_month}\\raw")
		if not file_path.exists():
			raise NotInCacheError("The requested data is currently not cached")
		index = next((i for i in file_path.iterdir() if i.is_file() and "contract" in i.name))

		if not index:
			raise NotInCacheError("Cannot find correct index file in local cache")
		index = pd.read_json(gzip.open(index))

		# Transform
		index['delivery_start'] = pd.to_datetime(index['delivery_start'])
		index['delivery_end'] = pd.to_datetime(index['delivery_end'])
		index = index.astype({"contract_id": "str", "delivery_areas": "str"})

		# Construct list of contract times
		contract_times = [tr[0]]
		[contract_times.append(contract_times[-1] + timedelta(minutes=ct)) for _ in
		 range(1, ((tr[1] - tr[0]) / (ct * 60)).seconds)]
		contract_times = [str(i.replace(tzinfo=timezone.utc)) for i in contract_times]

		# Filter dataframe down
		index = index.loc[(index._product.isin(PRODUCTS[contract_time])) & (index._product.isin(products)) & (
			index.delivery_areas.str.contains(client.delivery_area))]
		index = index.loc[index.delivery_start.isin(contract_times)]

		# Block product filter
		if not contract_time == "all" or not allow_udc:
			if client.exchange not in SYSTEMS["NordPool"]:
				index = index.loc[index.undrlng_contracts.isna()]
			else:
				index = index.loc[index.productType != "CUSTOM_BLOCK"]

		[contracts[
			f'{datetime.strftime(v.delivery_start, DATE_YMD_TIME_HM)} - {datetime.strftime(v.delivery_end, TIME_HM if v.delivery_start.date() == v.delivery_end.date() else DATE_YMD_TIME_HM)}'].append(
			v.contract_id) for k, v in index.iterrows()]

	return dict(contracts)



import json
from abc import ABC, abstractmethod
from decimal import Decimal
from datetime import datetime, timezone
from pathlib import Path
from typing import Union

import pandas as pd

from powerbot_backtesting.utils import _order_matching, _find_cache
from powerbot_backtesting.utils.constants import *


def generate_input_file(orderbooks):
	"""
	Generates a csv file to put positions and signals into to use with the BacktestingAlgo

	Args:
		orderbooks (dict{key: DataFrame}): Dictionary of order books

	Returns:
		csv file
	"""
	# File creation
	input_file = pd.DataFrame({"contract_id": [*orderbooks]})
	input_file["position"] = ""
	input_file.set_index("contract_id", inplace=True)

	# Caching
	cache_path = _find_cache().joinpath("analysis_input")
	cache_path.mkdir(parents=True, exist_ok=True)

	# File name
	f_count = 1
	while cache_path.joinpath(f"backtesting_input_{f_count}.csv").exists():
		f_count += 1
	input_file.to_csv(cache_path.joinpath(f"backtesting_input_{f_count}.csv"), sep=";")


class BacktestingAlgo(ABC):
	"""
	Class aiming to provide a backbone to backtesting strategies.

	This class is used by providing two parameters: the order books and an input file based on these order books.
	The input file can be generated with the command "generate_input_file". The output is a csv file that needs to be
	filled with positions and other signals you wish to utilize in your backtesting strategy.

	Any keyworded arguments you pass will be automatically added as class attributes. This allows you to only overwrite
	a single function and still be able to use all the parameters you need for your strategy.

	To write your own backtesting algorithm, create a new class that inherits from BacktestingAlgo and overwrite
	the algorithm function. All signals/ columns that you have added to the input file will be reflected as dictionary
	keys in self.params. Please be aware, that the column names will be the key names after conversion.

	The most important feature of this class is it's order matching logic. When writing your own algorithm, this
	function needs to be called whenever your calculations are done and you want to see if you would make any trades.

	Any trades you perform alter the order book and will also be gathered in a list of results.
	"""

	def __init__(self,
				 orderbooks: dict[str, pd.DataFrame],
				 file_input: Union[str, Path],
				 generate_report: bool = True,
				 **kwargs):
		self.orderbooks = orderbooks
		self.filename = file_input if isinstance(file_input, str) else str(file_input).split("\\")[-1]
		self.trade_list = {key: {} for key in [*self.orderbooks]}
		self.results = {key: {} for key in [*self.orderbooks]}
		self.generate_report = generate_report

		# Load file input
		cache_path = _find_cache().joinpath("analysis_input")

		try:
			self.algo_params = pd.read_csv(f"{cache_path}/{file_input}", sep=";")
		except FileNotFoundError:
			self.algo_params = pd.read_csv(file_input, sep=";")

		self.algo_params = self.algo_params.set_index("contract_id")

		# Convert all keyworded args to attributes
		[setattr(self, key, value) for key, value in kwargs.items()]

	def match_orders(self,
					 side: str,
					 orderbook: pd.DataFrame,
					 timestamp: str,
					 price: Union[int, float],
					 position: Union[int, float],
					 contract_time: int,
					 key: str,
					 order_execution: str) -> Decimal:
		"""
		Wrapper for _order_matching function.

		Matches orders according to input parameters; adds trades made to trade_list and returns the remaining quantity.

		The order_execution parameter can be added to decide according to which logic the quantity should be filled. Allowed
		values are:

		NON - No restriction, partial execution is allowed

		FOK - Fill or Kill, if order isn't filled completely by first matching order, next matching order is loaded ->
		if none match next order book is loaded

		IOC - Immediate and Cancel, order is executed to maximum extent by first matching order, next order book is loaded ->
		allows price adjustments

		Args:
			side (str): buy/sell
			orderbook (DataFrame): Single order book
			timestamp (str): Timestamp of order book
			price (int): Minimum/ Maximum Price for Transaction
			position (int, float): Quantity that can/ should be traded
			contract_time (int): contract time in minutes, either 60, 30 or 15
			key (str): Position in trade list
			order_execution (str): Type of order execution that should be simulated

		Returns:
			Decimal: remaining quantity
		"""
		return _order_matching(side, orderbook, timestamp, price, position, self.exec_orders_list, self.trade_list[key],
							   contract_time, order_execution)


	def generate_trading_report(self):
		"""
		Generates a report containing all trades made including prices, quantity and timestamps.

		Returns:
			bool
		"""
		cache_path = _find_cache().joinpath("analysis_output")
		cache_path.mkdir(parents=True, exist_ok=True)

		# File creation
		with open(cache_path.joinpath(self.filename.replace("input", "output").replace("csv", "json")), "w") as out:
			json.dump(self.results, out)

		return True

	def run(self) -> Union[dict, bool]:
		"""
		Main function to execute custom trading logic. Handles all necessary steps and then calls the custom algorithm
		function.

		Returns:
			dict or bool: generates json trading report or returns contents as dict
		"""

		for nr, (key, contract) in enumerate(self.orderbooks.items()):
			if isinstance(contract, pd.DataFrame) and contract.empty or not contract:
				self.results[key]["Remaining Quantity"] = "No data"
				continue

			self.exec_orders_list = {}

			self.params = self.algo_params.loc[key].to_dict()

			contract_time = key.split(" ")
			self.params["contract_time"] = int(
				(datetime.strptime(contract_time[3], TIME_HM) - datetime.strptime(contract_time[1],
																				  TIME_HM)).seconds / 60)
			# Overrideable function to set custom calculation logic on contract level
			self.adjust_params()

			# Iteration through every timestep/ orderbook
			for timestamp, orderbook in contract.items():
				# Overrideable function to set behaviour on order book level
				if not self.algorithm(timestamp, orderbook, key):
					break

			# If quantity remains
			if self.params["position"]:
				self.results[key]["Remaining Quantity"] = round(float(self.params["position"]), 1)

		# Create summary
		for contract, trades in self.trade_list.items():
			self.results[contract]["Initial Position"] = int(self.algo_params.loc[contract].position)
			self.results[contract]["Total Cash"] = round(sum([value["Cash"] for value in trades.values()]), 2)
			self.results[contract]["Trades"] = trades

		if self.generate_report:
			return self.generate_trading_report()

		return self.results

	def adjust_params(self):
		"""
		Overridable function that defines calculation logic on contract level. Any parameters that will be changed in
		this function will be usable for each iteration of the order books.

		This function should only be used to adjust the parameters in the member variable self.params. Please be aware,
		that deletion of the key "position" will lead to an error.
		"""
		return None

	@abstractmethod
	def algorithm(self,
				  timestamp: str,
				  orderbook: pd.DataFrame,
				  key: str) -> bool:
		"""
		Overridable function that defines behaviour and calculations for specific strategy. Please make sure to define
		the same input parameters as the original function.

		Since the order matching function already saves every important action it takes, it does not need any manual
		adjustments.

		The only requirement for this function is that at least the position has to be recalculated according to the
		return value of the order matching function. Every other parameter (e.g. price) can be changed according to the
		specific trading strategy.

		An example (position closer) has been defined here just to showcase how the function is supposed to work.
		Please be aware that all of the following logic will be overwritten by your own implementation.

		Args:
			timestamp (str): Timestamp of order book
			orderbook (pd.DataFrame): Order book
			key (str): Contract timeframe

		Returns:
			bool
		"""
		# Buy
		if self.params["position"] < 0:
			self.params["position"] = - self.match_orders(
				side="buy",
				orderbook=orderbook,
				timestamp=timestamp,
				price=self.params.get("price", 40) - 2,
				position=abs(self.params["position"]),
				contract_time=self.params["contract_time"],
				key=key,
				order_execution="NON")
			return True

		# Sell
		elif self.params["position"] > 0:
			self.params["position"] = self.match_orders(
				side="sell",
				orderbook=orderbook,
				timestamp=timestamp,
				price=self.params.get("price", 40) + 2,
				position=self.params["position"],
				contract_time=self.params["contract_time"],
				key=key,
				order_execution="NON")
			return True

		else:
			return False


def flexpos_algo(orderbooks: dict[str, pd.DataFrame],
                 marg_prices: list[float],
                 buy_sell_spread: list[float],
                 init_prod: list[float],
                 min_prod: list[float],
                 max_prod: list[float],
                 closing_from: int = 2,
                 marg_prices_pc: list[list[float]] = None) -> dict[str, dict]:
	"""
	Algorithm that connects a flexibility algorithm with a position closer. At the defined timestamp, the flexibility
	algo is replaced with the position closer algo, which tries to close a position with the provided information.

	Timestamps are not directly passed as arguments, but rather inferred by the closing_from parameter, which calculates
	the respective start times for the position closer based on the closing_from value and the delivery start of the
	contract.

	Args:
		orderbooks (dict{key: DataFrame}): Dictionary of Orderbooks
		marg_prices (list[float]): Marginal Price to fluctuate around/ close position
		buy_sell_spread (list[float]): Minimum Profit per MW
		init_prod (list[float]): Initial Quantity of produced MW
		min_prod (list[float]): Minimum Quantity of produced MW
		max_prod (list[float]): Maximum Quantity of produced MW
		closing_from (int): Amount of hours before contract delivery start algorithm should stop running at
		marg_prices_pc (list[list[float]]): Marginal prices specific for position closer (optional)

	Returns:
		dict[key: dict[Production_quant (float), Remaining_quant (float), Profit (float),
			Trades_Flex (list), Trades_PC (list)]]
	"""

	# Checking Parameters
	if not all(len(i) == len([*orderbooks]) for i in [marg_prices, buy_sell_spread, init_prod, min_prod, max_prod]):
		raise Exception(f"All parameters have to be same length as amount of contracts in orderbook!")

	# Executing Flex Algo
	results_flex = flex_algo(orderbooks, marg_prices, buy_sell_spread, init_prod, min_prod, max_prod, closing_from)

	# Setup for PC Algo
	positions = [v["Production"] for v in results_flex.values()]
	start_from = [datetime.strptime(i.replace(f"{str(i).split(' - ')[1]}", "").replace(" - ", ":00"), DATE_YMD_TIME_HMS)
	              for i in [*orderbooks]]
	orderbooks_pc = {key: {} for key in [*orderbooks]}

	for nr, (timestamp, contract) in enumerate(orderbooks.items()):
		for key, orderbook in contract.items():
			if datetime.strptime(key, DATE_YMD_TIME_HMS_TZ) >= start_from[nr].replace(tzinfo=timezone.utc):
				orderbooks_pc[timestamp][key] = orderbook
	marg_prices_pc = [[i, i] for i in marg_prices] if not marg_prices_pc else marg_prices_pc

	# Executing PC Algo
	results_pc = pc_algo(orderbooks_pc, positions, marg_prices_pc)

	# Summing up Results
	results_total = {
		key: {"Initial_quant": init_prod[nr], "Flex_quant": results_flex[key]["Production"], "Remaining_quant": 0,
		      "Profit": results_flex[key]["Profit"], "Trades_Flex": results_flex[key]["Trades"]} for nr, key in
		enumerate([*results_flex])}

	for key, contract in results_pc.items():
		for name, value in contract.items():
			if name == "Remaining_quant":
				results_total[key]["Remaining_quant"] = value
			if name == "Cash":
				results_total[key]["Profit"] = results_total[key]["Profit"] + Decimal(value)
			if name == "Trades":
				results_total[key]["Trades_PC"] = value

	return results_total


def flex_algo(orderbooks: dict[str, pd.DataFrame],
              marg_prices: list[float],
              buy_sell_spread: list[float],
              init_prod: list[float],
              min_prod: list[float],
              max_prod: list[float],
              closing_from: int = None) -> dict[str, dict]:
	"""
	Standard flexibility algorithm that aims to maximize profit by fluctuating around a marginal price.
	All parameters have to be given as lists of the same length as the amount of contracts in orderbooks.

	Args:
		orderbooks (dict{key: DataFrame}): Dictionary of Orderbooks
		marg_prices (list): Marginal Price to fluctuate around
		buy_sell_spread (list): Minimum Profit per KWh
		init_prod (list): Initial Quantity of produced MW
		min_prod (list): Minimum Quantity of produced MW
		max_prod (list): Maximum Quantity of produced MW
		closing_from (int): Amount of hours before contract delivery start algorithm should stop running at
			(only relevant for flexpos_algo)

	Returns:
		dict[key: dict[Production (float), Profit (float), Trades (list)]]
	"""
	# Setup
	trade_list = {key: {} for key in [*orderbooks]}
	results = {key: {} for key in [*orderbooks]}
	closing_timestamps = [
		datetime.strptime(i.replace(f"{str(i).split(' - ')[1]}", "").replace(" - ", ":00"), DATE_YMD_TIME_HMS) for i
		in [*orderbooks]] if closing_from else None

	for nr, (key, contract) in enumerate(orderbooks.items()):
		exec_orders_list = {}
		curr_prod = init_prod[nr]
		contract_time = key.split(" ")
		contract_time = int((datetime.strptime(contract_time[3], TIME_HM) - datetime.strptime(contract_time[1],
		                                                                                      TIME_HM)).seconds / 60)

		for timestamp, orderbook in contract.items():
			# Stopping Loop on closing Timestamp
			if closing_from and datetime.strptime(timestamp, DATE_YMD_TIME_HMS_TZ) >= closing_timestamps[nr].replace(tzinfo=timezone.utc):
				break
			# Buy
			if round(max_prod[nr] - curr_prod, 1) > 0.1:
				rem_quant_buy = _order_matching("buy", orderbook, timestamp, marg_prices[nr] - buy_sell_spread[nr],
				                                max_prod[nr] - curr_prod, exec_orders_list, trade_list[key],
				                                contract_time)

				curr_prod = max_prod[nr] - rem_quant_buy  # If Remaining Quantity. Subtract From Maximum

			# Sell
			if round(curr_prod - min_prod[nr], 1) > 0.1:
				rem_quant_sell = _order_matching("sell", orderbook, timestamp, marg_prices[nr] + buy_sell_spread[nr],
				                                 curr_prod - min_prod[nr], exec_orders_list, trade_list[key],
				                                 contract_time)

				curr_prod = min_prod[nr] + rem_quant_sell  # If Remaining Quantity, Add To Minimum

		cash = Decimal(sum([value["Cash"] for value in trade_list[key].values()]))  # Calculating Cash

		results[key]["Production"] = curr_prod
		results[key]["Profit"] = round(curr_prod * marg_prices[nr] - abs(cash), 2)
		results[key]["Trades"] = trade_list[key]

	return results


def pc_algo(orderbooks: dict[str, pd.DataFrame],
            positions: list[float],
            marg_prices: list[list[int]] = None) -> dict[str, dict]:
	"""
	Standard position closing algorithm that aims to close an open position, optionally taking a marginal price into
	consideration. Positional values have to match the amount of keys in the orderbooks in length (1 position per contract).

	Things to consider:
	This algorithm is best used with either a single orderbook for a certain timestamp that a position should be closed
	at or multiple orderbooks beginning from this timestamp. This simulates closing the position at the desired time.
	Alternatively, the flexpos_algo can be used to try to maximize profit before the position is closed.

	Args:
		orderbooks (dict{key: DataFrame}): Dictionary of Orderbooks
		positions (list): Positional values in MW (1 position per contract)
		marg_prices (list): List with 2 values representing the minimum amount an order can be matched with (bought or sold)

	Returns:
		dict[key: dict[Initial_quant (int/float), Remaining_quant (float), Cash (float), Trades (list)]]
	"""
	# Orderbooks dict object consists of keys for each contract with the aggregated orderbooks as dataframes attached

	# Checking Values
	if not isinstance(positions, list) or len(positions) != len([*orderbooks]):
		raise Exception("Parameter positions has to be list of floats of same length as contracts in orderbook!")

	if marg_prices:
		if not isinstance(marg_prices, list) or not all(len(i) == 2 for i in marg_prices):
			raise Exception("Marginal prices have to be provided as a list, containing 2 floats/integers representing"
			                " buy price and sell price!")
	else:
		marg_prices = [[1000, 0] for _ in [*orderbooks]]

	# Setup
	exec_orders_list = {}
	trade_list = {key: {} for key in [*orderbooks]}
	results = {key: {"Initial_quant": positions[nr], "Remaining_quant": 0, "Cash": 0, "Trades": []} for nr, key in
	           enumerate([*orderbooks])}

	for nr, (key, contract) in enumerate(orderbooks.items()):
		pos = positions[nr]
		buy_price = marg_prices[nr][0]
		sell_price = marg_prices[nr][1]
		contract_time = key.split(" ")
		contract_time = int((datetime.strptime(contract_time[3], TIME_HM) - datetime.strptime(contract_time[1],
		                                                                                      TIME_HM)).seconds / 60)

		if isinstance(contract, pd.DataFrame) and contract.empty or not contract:
			results[key]["Remaining_quant"] = "No data"
			continue

		for timestamp, orderbook in contract.items():
			# Buy
			if pos < 0:
				pos = -(_order_matching("buy", orderbook, timestamp, buy_price, abs(pos), exec_orders_list,
				                        trade_list[key], contract_time))

			# Sell
			elif pos > 0:
				pos = _order_matching("sell", orderbook, timestamp, sell_price, pos, exec_orders_list, trade_list[key],
				                      contract_time)

			else:
				break
		if pos:
			results[key]["Remaining_quant"] = round(pos, 1)

	for contract, trades in trade_list.items():
		results[contract]["Cash"] = round(sum([value["Cash"] for value in trades.values()]), 2)
		results[contract]["Trades"] = trades

	return results




from powerbot_backtesting.data_acquisition import init_client, get_contract_ids, get_public_trades, \
	get_public_trades_by_days, \
	get_contract_history, get_signals, get_own_trades, get_internal_trades, get_own_orders
from powerbot_backtesting.data_analysis import flex_algo, pc_algo, flexpos_algo, generate_input_file, BacktestingAlgo
from powerbot_backtesting.data_processing import get_orders, get_ohlc_data, get_orderbooks, calc_trade_vwap, \
	calc_single_vwap
from powerbot_backtesting.data_visualization import plot_ohlc, ohlc_table, plot_orderbook, plot_volume_history
from powerbot_backtesting.historic_data_acquisition import get_historic_data, get_history_key_info, \
	init_historic_client, get_historic_contract_ids
from powerbot_backtesting.historic_data_processing import process_historic_data
from powerbot_backtesting.models.exporter_models import ApiExporter, HistoryExporter, SQLExporter


import json
from collections import defaultdict
from datetime import datetime, timedelta
from time import sleep
from typing import Union

import pandas as pd
import requests
from dateutil.tz import tzutc
from powerbot_client import Configuration, ApiClient, ContractApi, Signal, Trade, \
	InternalTrade, OwnOrder

from powerbot_backtesting.historic_data_acquisition import get_historic_contract_ids
from powerbot_backtesting.models import HistoryApiClient
from powerbot_backtesting.utils import _cache_data, _check_contracts, _get_file_cachepath, _find_cache, \
	_get_private_data
from powerbot_backtesting.utils.constants import *


def init_client(api_key: str, host: str) -> ApiClient:
	"""
	Initializes PowerBot Client to enable data requests by the API.

	Args:
		api_key (str): API Key for PowerBot
		host (str): Host URL for PowerBot

	Returns:
		PowerBot ApiClient Object
	"""
	config = Configuration(api_key={'api_key_security': api_key}, host=host)
	return ApiClient(config)


def get_contract_ids(api_client: Union[ApiClient, HistoryApiClient],
                     time_from: datetime = None,
                     time_till: datetime = None,
                     contract_ids: list[str] = None,
                     contract_time: str = "all",
                     products: list[str] = None,
					 allow_udc: bool = False,
                     delivery_areas: list[str] = None,
                     return_contract_objects: bool = False) -> Union[dict[str, list[str]], tuple[dict, dict]]:
	"""
	Loads all contract IDs for specified timeframe, either by hours, single day or multiple days. Alternatively, a list
	of contract IDs can be passed instead of a timeframe, returning a dictionary of contract IDs compatible with all other
	functions of the Backtesting package. If return_contract_objects is True, a dictionary of contract objects will be
	returned as well.

	If a historic API client is used all parameters except time_from, time_till, contract_time, products and
	allow_udc will be ignored.

	If time_from includes hh:mm:ss
		loads data between both times in timesteps according to contract_time

	If time_from doesn't include hh:mm:ss
		loads data between both dates in timesteps according to contract_time

	Args:
		api_client: PowerBot ApiClient
		time_from (datetime): yyyy-mm-dd hh:mm:ss
		time_till (datetime): yyyy-mm-dd hh:mm:ss
		contract_ids (list[str]): Optionally, a list of specific contract IDs can be passed to return contract objects
		contract_time (str): all, hourly, half-hourly or quarter-hourly (all includes UDC)
		products (list): Optional list of specific products to return
		allow_udc (bool): True if user-defined contracts (block products) should also be returned on contract_time = all
		delivery_areas (list): List of EIC-codes
		return_contract_objects (bool): If True, returns complete Contract object

	Returns:
		dict{key: (list[str])}: Dictionary of Contract IDs
		(Optional) Contract: Contract Object
	"""
	# Check date input
	timeframes = {"all": 15, "hourly": 60, "half-hourly": 30, "quarter-hourly": 15}

	if not contract_ids:
		if not time_from or not time_till:
			raise TypeError("If no specific contract IDs are given, a time period has to be defined.")

		if not isinstance(time_from, datetime):
			raise ValueError("Please use datetime format (yyyy-mm-dd hh:mm:ss)")

	# Loading from historic data collection
	if isinstance(api_client, HistoryApiClient):
		if not time_from:
			raise AttributeError("Missing input parameter: time_from")
		return get_historic_contract_ids(client=api_client,
		                                 time_from=time_from,
		                                 time_till=time_till,
		                                 contract_time=contract_time,
		                                 products=products if products else [],
										 allow_udc=allow_udc)

	contract_api = ContractApi(api_client)
	products = [] if not products else products
	delivery_areas = [] if not delivery_areas else delivery_areas

	# if minutes are entered wrong, raise exception
	for t in [time_from, time_till]:
		if t.minute not in [0, 15, 30, 45]:
			raise ValueError("Time period has to start and end with one of these values: 0, 15, 30, 45")

	# Load Contract IDs
	contracts = defaultdict(list)

	# Loading by time_from & time_till
	if not contract_ids:
		while time_from != time_till:
			if contract_time != "all":
				new_contracts = contract_api.find_contracts(delivery_start=time_from,
				                                            delivery_end=time_from + timedelta(
					                                            minutes=timeframes[contract_time]))
			else:
				new_contracts = contract_api.find_contracts(delivery_start=time_from)

			# Check validity and add to contracts
			[contracts[f"{time_from.strftime(DATE_YMD_TIME_HM)} - {c.delivery_end.strftime(TIME_HM)}"].append(c)
			 if c.delivery_start.date() == c.delivery_end.date()
			 else contracts[f"{time_from.strftime(DATE_YMD_TIME_HM)} - {c.delivery_end.strftime(DATE_YMD_TIME_HM)}"].append(c)
			 for c in new_contracts if _check_contracts(c, delivery_areas, products, allow_udc)]

			time_from += timedelta(minutes=timeframes[contract_time])

	# Loading by specific contract IDs
	else:
		new_contracts = contract_api.find_contracts(contract_id=contract_ids)
		for c in new_contracts:
			# Check validity and add to contracts
			if _check_contracts(c, delivery_areas, products, allow_udc):
				# Add to contracts
				contracts[
					f"{c.delivery_start.strftime(DATE_YMD_TIME_HM)} - {c.delivery_end.strftime(DATE_YMD_TIME_HM)}"] \
					.append(c)

	# Cleanup
	contracts = {k: v for k, v in contracts.items() if v}
	contract_ids = {k: [c.contract_id for c in v] for k, v in contracts.items() if v}
	if not contracts:
		print("There was no contract data for your request!")

	# Returns Contract IDs and Object As A Whole
	if return_contract_objects:
		return contract_ids, contracts

	return contract_ids


def get_public_trades(api_client: Union[ApiClient, HistoryApiClient],
                      contract_ids: dict[str, list[str]],
                      contract_time: str,
                      delivery_area: str = None,
                      iteration_delay: float = 0.4,
                      serialize_data: bool = True,
                      add_vwap: bool = False,
                      use_cached_data: bool = True,
                      caching: bool = True,
                      gzip_files: bool = True,
                      as_csv: bool = False) -> dict[str, pd.DataFrame]:
	"""
	Load trade data for given contract IDs. If add_vwap is True, VWAP will be calculated for each trade, incorporating
	all previous trades.

	Args:
		api_client: PowerBot ApiClient if loading from API else HistoryApiClient
		contract_ids (dict): Dictionary of Contract IDs
		contract_time (str): all, hourly, half-hourly or quarter-hourly
		delivery_area (str): EIC Area Code for Delivery Area (not needed when loading from historic cache)
		iteration_delay (float): Optional delay between iterations to prevent hitting API rate limits
		serialize_data (bool): If False, request is received without serialization. Recommended for large data collections
		add_vwap (bool): If True, additional VWAP parameters will be added to each dataframe
		use_cached_data (bool): If True, function tries to load data from cache wherever possible
		caching (bool): True if data should be cached
		gzip_files (bool): True if cached files should be gzipped
		as_csv (bool): if True, will save files as CSV, additionally to JSON

	Returns:
		dict{key: DataFrame}: Dictionary of DataFrames
	"""
	# Setup
	trades = {}
	missing_contracts = {}
	delivery_area = api_client.delivery_area if isinstance(api_client, HistoryApiClient) else delivery_area

	# Load Data
	# Load from Cache
	if use_cached_data:
		# Find __cache__ directory
		cache_path = _find_cache()

		for key, value in contract_ids.items():
			filepath = _get_file_cachepath(api_client, key, delivery_area)
			tmp_df = None

			for i in [".json.gz", ".json"]:
				if cache_path.joinpath(f"{filepath}_trades{i}").exists():
					tmp_df = pd.read_json(cache_path.joinpath(f"{filepath}_trades{i}"), dtype=False)

			if isinstance(tmp_df, pd.DataFrame):
				tmp_df['api_timestamp'] = pd.to_datetime(tmp_df['api_timestamp'])
				tmp_df['exec_time'] = pd.to_datetime(tmp_df['exec_time'])
				tmp_df = tmp_df.astype({"price": "float64", "trade_id": "str", "contract_id": "str"})
				for i in ["price", "quantity"]:
					tmp_df[i] = round(tmp_df[i], 2)

				# Filter out any contract IDs that are not in value
				trades[key] = tmp_df.loc[tmp_df.contract_id.isin(value)]

			else:
				# Save Missing Contract IDs
				missing_contracts[key] = value

	# Historic data -> cache only
	if isinstance(api_client, ApiClient):
		contract_api = ContractApi(api_client)
		contract_ids = missing_contracts if use_cached_data else contract_ids

		for key, value in contract_ids.items():
			public_trade_history = []
			for nr, item in enumerate(value):
				from_public_trade = 0
				more_public_trades = True
				while more_public_trades:
					if serialize_data:
						public_trades = contract_api.get_public_trades(
							contract_id=item,
							delivery_area=delivery_area,
							offset=from_public_trade,
							limit=500
						)
						public_trade_history.extend([trade.to_dict() for trade in public_trades])

					else:
						endpoint = f"{api_client.configuration.host}/contract/{item}/{delivery_area}/publictrades?offset={from_public_trade}&limit=500"
						headers = {"accept": "application/json",
						           "api_key_security": api_client.configuration.api_key['api_key_security']}
						public_trades = json.loads(requests.get(endpoint, headers=headers).text)
						public_trade_history += public_trades

					if len(public_trades) < 500:
						more_public_trades = False
					from_public_trade += len(public_trades)

			if len(public_trade_history) == 0:
				continue

			df_trades = pd.DataFrame(public_trade_history).sort_values(by=['exec_time'], ascending=True)
			df_trades = df_trades.reset_index(drop=True)
			df_trades['api_timestamp'] = pd.to_datetime(df_trades['api_timestamp'], utc=True)
			df_trades['exec_time'] = pd.to_datetime(df_trades['exec_time'], utc=True)

			trades[key] = df_trades
			sleep(iteration_delay)

	if add_vwap:
		from powerbot_backtesting.data_processing import calc_trade_vwap
		for k, v in trades.items():
			trades[k] = calc_trade_vwap(api_client=api_client, trade_data={k: v}, contract_time=contract_time,
			                            delivery_area=delivery_area, index="all")

	if caching:
		_cache_data("trades", trades, delivery_area, api_client=api_client, gzip_files=gzip_files, as_csv=as_csv)
	return trades


def get_public_trades_by_days(api_client: ApiClient,
                              previous_days: int,
                              delivery_area: str,
                              time_from: datetime = None,
                              contract_time: str = None,
                              contract_id: str = None) -> dict[str, pd.DataFrame]:
	"""
	Gets the contract ID specified by a timeframe or directly by ID and load all trade data for this contract and all
	contracts in the same timeframe for X previous days.

	Args:
		api_client: PowerBot ApiClient
		time_from (str/ datetime): YYYY-MM-DD hh:mm:ss
		previous_days (int): Amount of previous days to load data for
		delivery_area (str): EIC Area Code for Delivery Area
		contract_time (str): hourly, half-hourly or quarter-hourly
		contract_id (str): specific contract ID

	Returns:
		dict{key: DataFrame}: Dictionary of DataFrames
	"""
	if not time_from and not contract_id:
		raise TypeError("Either time_from and contract_time or a specific contract ID have to be passed.")
	if time_from and not contract_time:
		raise TypeError("If time_from is given, contract_time has to be passed as well.")

	products = None
	use_cached_data = True
	timeframes = {"hourly": 60, "half-hourly": 30, "quarter-hourly": 15}

	try:
		if time_from:
			time_till = time_from + timedelta(minutes=timeframes[contract_time])

		else:
			contract = ContractApi(api_client).find_contracts(contract_id=[contract_id])[0]
			if contract.type == "UDC":
				raise TypeError("This function does not work for user defined contracts (UDC).")
			time_from = contract.delivery_start
			time_till = contract.delivery_end
			timeframes = {60: "hourly", 30: "half-hourly", 15: "quarter-hourly"}
			contract_time = timeframes[int((contract.delivery_end - contract.delivery_start).seconds / 60)]
			products = [contract.product]
			use_cached_data = False

		contract_ids = get_contract_ids(api_client=api_client, time_from=time_from, time_till=time_till,
		                                contract_time=contract_time, products=products, delivery_areas=[delivery_area])

		for _ in range(previous_days):
			time_from -= timedelta(days=1)
			time_till -= timedelta(days=1)

			contract_ids.update(get_contract_ids(api_client=api_client, time_from=time_from, time_till=time_till,
			                                     contract_time=contract_time, products=products))
	except (ValueError, TypeError):
		raise ValueError("Please use correct format: yyyy-mm-dd hh:mm:ss")

	# Get Trade Data
	return get_public_trades(api_client=api_client,
	                         contract_ids=contract_ids,
	                         delivery_area=delivery_area,
	                         contract_time=contract_time,
	                         use_cached_data=use_cached_data)


def get_contract_history(api_client: Union[ApiClient, HistoryApiClient],
                         contract_ids: dict[str, list[str]],
                         delivery_area: str = None,
                         iteration_delay: float = 0.4,
                         serialize_data: bool = True,
                         use_cached_data: bool = True,
                         caching: bool = True,
                         gzip_files: bool = True,
                         as_csv: bool = False) -> dict[str, pd.DataFrame]:
	"""
	Load contract history for given contract IDs.

	Args:
		api_client: PowerBot ApiClient if loading from API else HistoryApiClient
		contract_ids (dict): Dictionary of Contract IDs
		delivery_area (str): EIC Area Code for Delivery Area (not needed when loading from historic cache)
		iteration_delay (float): Optional delay between iterations to prevent hitting API rate limits
		serialize_data (bool): If False, request is received without serialization. Recommended for large data collections
		use_cached_data (bool): If True, function tries to load data from cache wherever possible
		caching (bool): True if data should be cached
		gzip_files (bool): True if cached files should be gzipped
		as_csv (bool): if True, will save files as CSV, additionally to JSON

	Returns:
		dict{key: DataFrame}: Dictionary of DataFrames
	"""
	# Setup
	orders = {}
	missing_contracts = {}
	delivery_area = api_client.delivery_area if isinstance(api_client, HistoryApiClient) else delivery_area

	if use_cached_data:
		# Find __cache__ directory
		cache_path = _find_cache()
		for key, value in contract_ids.items():
			filepath = _get_file_cachepath(api_client, key, delivery_area)
			tmp_df = None

			for i in [".json.gz", ".json"]:
				if cache_path.joinpath(f"{filepath}_ordhist{i}").exists():
					tmp_df = pd.read_json(cache_path.joinpath(f"{filepath}_ordhist{i}"), dtype=False,
					                      convert_dates=False)

			if isinstance(tmp_df, pd.DataFrame):
				tmp_df['as_of'] = pd.to_datetime(tmp_df['as_of'])

				cols = {"internal_trades": "object", "contract_id": "str", "auction_price": "float64"}
				cols = {k: v for k, v in cols.items() if k in tmp_df.columns}
				tmp_df = tmp_df.astype(cols, errors='ignore')
				for i in ["best_bid_price", "best_bid_quantity", "best_ask_price", "best_ask_quantity", "last_price",
				          "last_quantity", "total_quantity", "high", "low", "vwap"]:
					try:
						tmp_df[i] = round(tmp_df[i], 2)
					except (TypeError, KeyError):
						pass

				if "orders" in tmp_df.columns:
					order_list = tmp_df.orders.tolist()
					for order_type in ["bid", "ask"]:
						for i in order_list:
							if order_type in i and i[order_type]:
								for x in i[order_type]:
									for param in ["quantity", "price"]:
										x[param] = round(x[param], 2)
									try:
										x["order_entry_time"] = datetime.strptime(x["order_entry_time"],
										                                          "%Y-%m-%dT%H:%M:%S.%fZ").replace(
											tzinfo=tzutc())
									except ValueError:
										x["order_entry_time"] = datetime.strptime(x["order_entry_time"],
										                                          "%Y-%m-%dT%H:%M:%SZ").replace(
											microsecond=0, tzinfo=tzutc())
					tmp_df["orders"] = order_list

				# Filter out any contract IDs that are not in value
				orders[key] = tmp_df.loc[tmp_df.contract_id.isin(value)]

			else:
				# Save Missing Contract IDs
				missing_contracts[key] = value

	# Historic data -> cache only
	if isinstance(api_client, ApiClient):
		contract_api = ContractApi(api_client)
		contract_ids = missing_contracts if use_cached_data else contract_ids

		for key, value in contract_ids.items():
			public_contract_history = []
			for nr, item in enumerate(value):
				more_revisions = True
				from_revision = 0

				while more_revisions:
					if serialize_data:
						contract_history = contract_api.get_contract_history(
							contract_id=item,
							delivery_area=delivery_area,
							with_orders=True,
							with_owntrades=False,
							offset=from_revision,
							limit=150
						)
						public_contract_history.extend([trade.to_dict() for trade in contract_history])

					else:
						endpoint = f"{api_client.configuration.host}/contract/{item}/{delivery_area}/history?offset={from_revision}&limit=150&with_owntrades=false&with_signals=false&with_orders=true"
						headers = {"accept": "application/json",
						           "api_key_security": api_client.configuration.api_key['api_key_security']}
						contract_history = json.loads(requests.get(endpoint, headers=headers).text)
						public_contract_history += contract_history

					if len(contract_history) < 150:
						more_revisions = False
					from_revision += len(contract_history)

			if len(public_contract_history) == 0:
				continue

			df_history = pd.DataFrame(public_contract_history).sort_values(by=['as_of'], ascending=True)
			df_history = df_history.reset_index(drop=True)
			df_history['as_of'] = pd.to_datetime(df_history['as_of'], utc=True)
			df_history.drop(columns=["auction_price", "internal_trades"], inplace=True, errors="ignore")

			orders[key] = df_history
			sleep(iteration_delay)

	if caching:
		_cache_data("ordhist", orders, delivery_area, api_client=api_client, gzip_files=gzip_files, as_csv=as_csv)
	return orders


def get_signals(api_client: ApiClient,
                time_from: datetime,
                time_till: datetime,
                delivery_area: str = None,
                portfolio_id: list[str] = None) -> list[Signal]:
	"""
	Function gathers all Signals received by the API in the specified time period and gathers them in a list.

	Args:
		api_client: PowerBot ApiClient
		time_from (datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
		time_till (datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
		delivery_area (str): EIC Area Code for Delivery Area
		portfolio_id (str): List of all portfolios that signals should be loaded from

	Returns:
		list[Signal]
	"""
	if not isinstance(time_from, datetime) or not isinstance(time_till, datetime):
		raise TypeError("time_from and time_till need to be of type datetime")

	return _get_private_data(api_client=api_client,
	                         data_type="signal",
	                         time_from=time_from,
	                         time_till=time_till,
	                         delivery_area=delivery_area,
	                         portfolio_id=portfolio_id)


def get_own_trades(api_client: ApiClient,
                   time_from: datetime,
                   time_till: datetime,
                   delivery_area: str = None,
                   portfolio_id: list[str] = None) -> list[Trade]:
	"""
	Function to collect all Own Trades for the defined time period, either specific to portfolio and/or delivery area
	or all portfolios and delivery areas used API key has access to.

	Args:
		api_client: PowerBot ApiClient
		time_from (str/ datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
		time_till (str/ datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
		delivery_area (str): EIC Area Code for Delivery Area
		portfolio_id (list[str]): List of specific portfolio IDs to load trades from. If left out, will load from all IDs.

	Returns:
		list[Trade]
	"""
	if not isinstance(time_from, datetime) or not isinstance(time_till, datetime):
		raise TypeError("time_from and time_till need to be of type datetime")

	return _get_private_data(api_client=api_client,
	                         data_type="own_trade",
	                         time_from=time_from,
	                         time_till=time_till,
	                         delivery_area=delivery_area,
	                         portfolio_id=portfolio_id)


def get_internal_trades(api_client: ApiClient,
                        time_from: datetime,
                        time_till: datetime,
                        delivery_area: str = None,
                        portfolio_id: list[str] = None) -> list[InternalTrade]:
	"""
	Function to collect all Internal Trades for the defined time period, either specific to portfolio and/or delivery
	area or all portfolios and delivery areas used API key has access to.

	Args:
		api_client: PowerBot ApiClient
		time_from (str/ datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
		time_till (str/ datetime): YYYY-MM-DD or YYYY-MM-DD hh:mm:ss
		delivery_area (str): EIC Area Code for Delivery Area
		portfolio_id (list[str]): List of specific portfolio IDs to load trades from. If left out, will load from all IDs.

	Returns:
		list[InternalTrade]
	"""
	if not isinstance(time_from, datetime) or not isinstance(time_till, datetime):
		raise TypeError("time_from and time_till need to be of type datetime")

	return _get_private_data(api_client=api_client,
	                         data_type="internal_trade",
	                         time_from=time_from,
	                         time_till=time_till,
	                         delivery_area=delivery_area,
	                         portfolio_id=portfolio_id)


def get_own_orders(api_client: ApiClient,
                   delivery_area: str = None,
                   portfolio_id: list[str] = None,
                   active_only: bool = False) -> list[OwnOrder]:
	"""
	Function to collect all available Own Orders, either specific to portfolio and/or delivery area or all portfolios
	and delivery areas used API key has access to.

	Args:
		api_client: PowerBot ApiClient
		delivery_area (str): EIC Area Code for Delivery Area
		portfolio_id (list[str]): List of specific portfolio IDs to load trades from. If left out, will load from all IDs.
		active_only (bool):  True if only active orders should be loaded. If False, loads also hibernate and inactive.

	Returns:
		list[OwnOrder]
	"""
	return _get_private_data(api_client=api_client,
	                         data_type="own_order",
	                         delivery_area=delivery_area,
	                         portfolio_id=portfolio_id,
	                         active_only=active_only)
